"""Databases.py: This file contains MINE database classes including database
loading and writing functions."""
import datetime
import multiprocessing
import os
import platform
import sys
from copy import deepcopy
from math import ceil
from shutil import move
from subprocess import call
from typing import List, Union

import pymongo
from pymongo.errors import ServerSelectionTimeoutError
from rdkit.Chem import AllChem
from rdkit.RDLogger import logger

from minedatabase import utils


# from minedatabase.NP_Score import npscorer as nps

# nps_model = nps.readNPModel()

lg = logger()
lg.setLevel(4)


def establish_db_client(uri: str = None) -> pymongo.MongoClient:
    """Establish a connection to a mongo database given a URI.

    Uses the provided URI to connect to a mongoDB. If none is given
    the default URI is used when using pymongo.

    Parameters
    ----------
    uri : str, optional
        URI to connect to mongo DB, by default None.

    Returns
    -------
    pymongo.MongoClient
        Connection to the specified mongo instance.

    Raises
    ------
    IOError
        Attempt to connect to database timed out.
    """
    # If a connection string is given use that, otherwise go to local
    try:
        if uri:
            client = pymongo.MongoClient(uri, ServerSelectionTimeoutMS=5)
        else:
            client = pymongo.MongoClient(ServerSelectionTimeoutMS=5)
    except ServerSelectionTimeoutError:
        raise IOError(
            "Failed to load database client. Please verify that " "mongod is running"
        )
    return client


class MINE:
    """
    This class provides an interface to the MongoDB and some useful functions.

    Parameters
    ----------
    name : str
        Name of the database to work with.
    uri : str, optional
        uri of the mongo server, by default "mongodb://localhost:27017/".

    Attributes
    ----------
    client : pymongo.MongoClient
        client connection to the MongoDB.
    compounds : Collection
        Compounds collection.
    core_compounds : Collection
        Core compounds collection.
    meta_data : Collection
        Metadata collection.
    models : Collection
        Models collection.
    name : str
        Name of the database
    operators : Collection
        Operators collection.
    reactions : Collection
        Reactions collection.
    target_compounds : Collection
        Target compounds collection.
    uri : str
        MongoDB connection string.
    """

    def __init__(self, name: str, uri: str = "mongodb://localhost:27017/") -> None:
        self.client = establish_db_client(uri)
        self.uri = uri
        self._db = self.client[name]
        self._core_db = self.client.core
        self.core_compounds = self._core_db.compounds
        self.name = name
        self.meta_data = self._db.meta_data
        self.compounds = self._db.compounds
        self.target_compounds = self._db.target_compounds
        self.reactions = self._db.reactions
        self.operators = self._db.operators
        self.models = self._db.models
        self.reactant_in = self._db.reactant_in
        self.product_of = self._db.product_of
        # self.nps_model = nps.readNPModel()
        self._mass_cache = {}  # for rapid calculation of reaction mass change

    def add_reaction_mass_change(self, reaction: str = None) -> Union[float, None]:
        """Calculate the change in mass between reactant and product compounds.

        This is useful for discovering compounds in molecular networking. If no reaction
        is specified then mass change of each reaction in the database will be
        calculated.

        Parameters
        ----------
        reaction : str, optional
            Reaction ID to calculate the mass change for, by default None.

        Returns
        -------
        float, optional
            Mass change of specified reaction. None if masses not all found.
        """

        def _get_mass(_id):
            if _id not in self._mass_cache:
                try:
                    self._mass_cache["_id"] = self.compounds.find_one(
                        {"_id": _id}, {"Mass": 1}
                    )["Mass"]
                except (TypeError, KeyError):
                    raise ValueError(
                        f"A mass value for {_id} was not found in "
                        "compounds collection"
                    )
            return self._mass_cache["_id"]

        def _get_mass_change(rxn):
            if isinstance(rxn["Reactants"][0], dict):
                key = "c_id"
            else:
                key = 1
            start_mass = _get_mass(rxn["Reactants"][0][key])
            return [
                _get_mass(product[key]) - start_mass
                for product in rxn["Products"]
                if product[key][0] == "C"
            ]

        if reaction:
            return _get_mass_change(reaction)

        bulk = self.reactions.initialize_unordered_bulk_op()
        reactions = self.reactions.find({"Mass_Change": {"$exists": 0}})
        for rxn in reactions:
            try:
                bulk.find({"_id": rxn["_id"]}).update(
                    {"$set": {"Mass_Change": _get_mass_change(rxn)}}
                )
            except ValueError as e:
                print(e.args[0])
                print(f"Failed to parse rxn {rxn['_id']}")
        bulk.execute()

    def generate_image_files(
        self,
        path: str,
        query: dict = None,
        dir_depth: int = 0,
        img_type: str = "svg:-a,nosource,w500,h500",
        convert_r: bool = False,
    ) -> None:
        """Generates image files for compounds in database using ChemAxon's MolConvert.

        Parameters
        ----------
        path : str
            Target directory for image file.
        query : dict, optional
            Query to limit number of files generated, by default None.
        dir_depth : int, optional
            The number of directory levels to split the compounds
            into for files system efficiency. Ranges from 0 (all in top
            level directory) to the length of the file name (40 for MINE hashes),
            by default 0.
        img_type : str, optional
            Type of image file to be generated. See molconvert
            documentation for valid options, by default 'svg:-a,nosource,w500,h500'.
        convert_r : bool, optional
            Convert R in the smiles to *, by default False.
        """
        ids = []
        extension = img_type.split(":")[0]
        structure_file = os.path.join(path, "tmp.smiles")
        if not query:
            query = {}

        if not os.path.exists(path):
            if sys.platform == "linux":
                os.system(f"sudo mkdir {path} -m 777")
            else:
                os.mkdir(path)
        with open(structure_file, "w") as outfile:
            for comp in self.compounds.find(query, {"SMILES": 1}):
                if convert_r:
                    outfile.write(f"{comp['SMILES'].replace('R', '*')}\n")
                else:
                    outfile.write(f"{comp['SMILES']}\n")
                ids.append(comp["_id"])
        if platform.system() == "Windows":
            rc = call(
                [f"molconvert -mo '{path}/.{extension}' {img_type} '{structure_file}'"],
                shell=True,
            )
        else:
            rc = call(
                [f"molconvert -mo '{path}/.{extension}' {img_type} '{structure_file}'"],
                shell=True,
            )
        if rc:
            raise RuntimeError(f"molconvert returned {rc}")
        os.remove(structure_file)

        for i, _id in enumerate(ids):
            old = os.path.join(path, f"{i+1}.{extension}")
            new = path
            for j in range(0, dir_depth):
                new = os.path.join(new, _id[j])
            if not os.path.exists(new):
                os.makedirs(new)
            new = os.path.join(new, _id + "." + extension)
            if os.path.isfile(old):
                move(old, new)

    def build_indexes(self) -> None:
        """Build indexes for efficient querying of the database."""
        self.core_compounds.drop_indexes()
        self.core_compounds.create_index([("Mass", pymongo.ASCENDING)])
        self.core_compounds.create_index("Inchikey")
        self.core_compounds.create_index("MINES")
        self.compounds.create_index("Inchikey")
        self.compounds.create_index("Inchi")
        self.compounds.create_index("SMILES")
        self.reactions.create_index("Reactants.c_id")
        self.reactions.create_index("Products.c_id")
        self.meta_data.insert_one(
            {"Timestamp": datetime.datetime.now(), "Action": "Database indexes built"}
        )


# Functions to write data to MINE
# Reactions
def write_reactions_to_mine(
    reactions: List[dict], db: MINE, chunk_size: int = 10000
) -> None:
    """Write reactions to reaction collection of MINE.

    Parameters
    ----------
    reactions : List[dict]
        Dictionary of reactions to write.
    db : MINE
        MINE object to write reactions with.
    chunk_size : int, optional
        Size of chunks to break reactions into when writing, by default 10000.
    """
    n_rxns = len(reactions)
    for i, rxn_chunk in enumerate(utils.Chunks(reactions, chunk_size)):
        if i % 20 == 0:
            print(f"Writing Reactions: Chunk {i} of {int(n_rxns/chunk_size) + 1}")
        rxn_requests = [
            pymongo.InsertOne(utils.convert_sets_to_lists(rxn_dict))
            for rxn_dict in rxn_chunk
        ]

        db.reactions.bulk_write(rxn_requests, ordered=False)


# Compounds
def write_compounds_to_mine(
    compounds: List[dict], db: MINE, chunk_size: int = 10000, processes: int = 1
) -> None:
    """Write compounds to reaction collection of MINE.

    Parameters
    ----------
    compounds : List[dict]
        Dictionary of compounds to write.
    db : MINE
        MINE object to write compounds with.
    chunk_size : int, optional
        Size of chunks to break compounds into when writing, by default 10000.
    processes : int, optional
        Number of processors to use, by default 1.
    """
    n_cpds = len(compounds)
    if processes == 1:
        pool = None
    else:
        pool = multiprocessing.Pool(processes)

    for i, cpd_chunk in enumerate(utils.Chunks(compounds, chunk_size)):
        if i % 20 == 0:
            print(f"Writing Compounds: Chunk {i} of {int(n_cpds/chunk_size) + 1}")

        cpd_requests = []
        reactant_in_requests = []
        product_of_requests = []

        if pool:
            for res in pool.imap_unordered(_get_cpd_insert, cpd_chunk):
                cpd_request, reactant_in_request, product_of_request = res
                cpd_requests.append(cpd_request)
                reactant_in_requests.extend(reactant_in_request)
                product_of_requests.extend(product_of_request)
        else:
            for res in map(_get_cpd_insert, cpd_chunk):
                cpd_request, reactant_in_request, product_of_request = res
                cpd_requests.append(cpd_request)
                reactant_in_requests.extend(reactant_in_request)
                product_of_requests.extend(product_of_request)

        db.compounds.bulk_write(cpd_requests, ordered=False)
        if reactant_in_requests:
            db.reactant_in.bulk_write(reactant_in_requests, ordered=False)
        if product_of_requests:
            db.product_of.bulk_write(product_of_requests, ordered=False)

    if pool:
        pool.close()
        pool.join()


def _get_cpd_insert(cpd_dict: dict):
    output_keys = [
        "_id",
        "ID",
        "SMILES",
        "InChI_key",
        "Type",
        "Generation",
        "Expand",
        "Reactant_in",
        "Product_of",
        "Matched_Peak_IDs",
        "Matched_Adducts",
        "Predicted_RT",
    ]

    # create Reactant_in
    reactant_in_requests = []
    product_of_requests = []
    insert_dict = {
        key: cpd_dict.get(key) for key in output_keys if cpd_dict.get(key) != None
    }
    if "Reactant_in" in insert_dict:
        chunked_reactant_in = _get_reactant_in_insert(cpd_dict)
        insert_dict["Reactant_in"] = []
        for r_in_dict in chunked_reactant_in:
            reactant_in_requests.append(pymongo.InsertOne(r_in_dict))
            insert_dict["Reactant_in"].append(r_in_dict["_id"])

    # create Product_of
    if "Product_of" in insert_dict:
        chunked_product_of = _get_product_of_insert(cpd_dict)
        insert_dict["Product_of"] = []
        for p_of_dict in chunked_product_of:
            product_of_requests.append(pymongo.InsertOne(p_of_dict))
            insert_dict["Product_of"].append(p_of_dict["_id"])

    cpd_request = pymongo.InsertOne(insert_dict)
    return cpd_request, reactant_in_requests, product_of_requests


def _get_reactant_in_insert(compound: dict) -> List[dict]:
    """Write reactants_in, ensuring memory size isn't too big.

    MongoDB only allows < 16 MB entries. This function breaks large reactants_in
    up to ensure this doesn't happen.

    Parameters
    ----------
    compounds : List[dict]
        Dictionary of compounds to write.
    db : MINE
        MINE object to write compounds with.

    Returns
    -------
    List[dict]
        dicts of reactant_in to insert
    """

    # Get number of chunks reactant_in must be broken up into
    # 16 MB is the max for BSON, cut to 14 MB max just to be safe
    # Also some weirdness is that max_size must be 1 order of magnitude lower
    max_size = 1.4 * 10 ** 6
    r_in_size = sys.getsizeof(compound["Reactant_in"])
    chunks, rem = divmod(r_in_size, max_size)

    if rem:
        chunks += 1
    chunk_size = ceil(len(compound["Reactant_in"]) / chunks)

    # Generate the InsertOne requests
    r_in_chunks = utils.Chunks(compound["Reactant_in"], chunk_size, return_list=True)

    requests = []
    for i, r_in_chunk in enumerate(r_in_chunks):
        requests.append(
            {
                "_id": f"{compound['_id']}_{i}",
                "c_id": compound["_id"],
                "Reactant_in": r_in_chunk,
            }
        )

    return requests


def _get_product_of_insert(compound: dict) -> List[dict]:
    """Write reactants_in, ensuring memory size isn't too big.

    MongoDB only allows < 16 MB entries. This function breaks large reactants_in
    up to ensure this doesn't happen.

    Parameters
    ----------
    compounds : List[dict]
        Dictionary of compounds to write.
    db : MINE
        MINE object to write compounds with.

    Returns
    -------
    List[dict]
        dicts of product_of to insert
    """

    # Get number of chunks product_of must be broken up into
    # 16 MB is the max for BSON, cut to 14 MB max just to be safe
    # Also some weirdness is that max_size must be 1 order of magnitude lower
    max_size = 1.4 * 10 ** 6
    p_of_size = sys.getsizeof(compound["Product_of"])
    chunks, rem = divmod(p_of_size, max_size)
    if rem:
        chunks += 1
    chunk_size = ceil(len(compound["Product_of"]) / chunks)

    # Generate the InsertOne requests
    p_of_chunks = utils.Chunks(compound["Product_of"], chunk_size, return_list=True)

    requests = []
    for i, p_of_chunk in enumerate(p_of_chunks):
        requests.append(
            {
                "_id": f"{compound['_id']}_{i}",
                "c_id": compound["_id"],
                "Product_of": p_of_chunk,
            }
        )

    return requests


# Core Compounds
def write_core_compounds(
    compounds: List[dict], db: MINE, mine: str, chunk_size: int = 10000, processes=1
) -> None:
    """Write core compounds to the core compound database.

    Calculates and formats compounds into appropriate form to insert into the
    core compound database in the mongo instance. Core compounds are attempted
    to be inserted and collisions are detected on the database. The list of
    MINEs a given compound is found in is updated as well.

    Parameters
    ----------
    compounds : dict
        List of compound dictionaries to write.
    db : MINE
        MINE object to write core compounds with.
    mine : str
        Name of the MINE.
    chunk_size : int, optional
        Size of chunks to break compounds into when writing, by default 10000.
    processes : int, optional
        The number of processors to use, by default 1.
    """
    n_cpds = len(compounds)
    if processes == 1:
        pool = None
    else:
        pool = multiprocessing.Pool(processes)

    for i, cpd_chunk in enumerate(utils.Chunks(compounds, chunk_size)):
        if i % 20 == 0:
            print(f"Writing Compounds: Chunk {i} of {int(n_cpds/chunk_size) + 1}")

        # Capture annoying RDKit output

        cpd_chunk = [deepcopy(cpd) for cpd in cpd_chunk if cpd["_id"].startswith("C")]
        if pool:
            core_requests = [req for req in pool.map(_get_core_cpd_insert, cpd_chunk)]
        else:
            core_requests = [req for req in map(_get_core_cpd_insert, cpd_chunk)]

        core_update_requests = [
            _get_core_cpd_update(cpd_dict, mine) for cpd_dict in cpd_chunk
        ]

        # Need to write update (i.e. what keeps track of which mines the
        # compound has been in)
        # first to ensure cpd exists
        db.core_compounds.bulk_write(core_requests)
        db.core_compounds.bulk_write(core_update_requests)
    if pool:
        pool.close()
        pool.join()


def _get_core_cpd_update(cpd_dict: dict, mine: str) -> pymongo.UpdateOne:
    return pymongo.UpdateOne({"_id": cpd_dict["_id"]}, {"$addToSet": {"MINES": mine}})


def _get_core_cpd_insert(cpd_dict: dict) -> pymongo.UpdateOne:
    """Generate core compound to be inserted"""
    core_keys = ["_id", "SMILES", "Inchi", "InchiKey", "Mass", "Formula"]
    core_dict = {
        key: cpd_dict.get(key) for key in core_keys if cpd_dict.get(key) != None
    }

    mol_object = AllChem.MolFromSmiles(core_dict["SMILES"])
    rdk_fp = [
        i
        for i, val in enumerate(list(AllChem.RDKFingerprint(mol_object, fpSize=512)))
        if val
    ]

    # Store all different representations of the molecule (SMILES, Formula,
    #  InChI key, etc.) as well as its properties in a dictionary
    if not "SMILES" in core_dict:
        core_dict["SMILES"] = AllChem.MolToSmiles(mol_object, True)
    if not "Inchi" in core_dict:
        core_dict["Inchi"] = AllChem.MolToInchi(mol_object)
    if not "Inchikey" in core_dict:
        core_dict["Inchikey"] = AllChem.InchiToInchiKey(core_dict["Inchi"])

    core_dict["Mass"] = AllChem.CalcExactMolWt(mol_object)
    core_dict["Charge"] = AllChem.GetFormalCharge(mol_object)
    core_dict["Formula"] = AllChem.CalcMolFormula(mol_object)
    core_dict["logP"] = AllChem.CalcCrippenDescriptors(mol_object)[0]
    core_dict["RDKit_fp"] = rdk_fp
    core_dict["len_RDKit_fp"] = len(rdk_fp)
    # core_dict['NP_likeness'] = nps.scoreMol(mol_object, nps_model)
    core_dict["Spectra"] = {}
    # Record which expansion it's coming from
    core_dict["MINES"] = []

    return pymongo.UpdateOne(
        {"_id": core_dict["_id"]}, {"$setOnInsert": core_dict}, upsert=True
    )


# Target Compounds
def write_targets_to_mine(
    targets: List[dict], db: MINE, chunk_size: int = 10000
) -> None:
    """Write target compounds to target collection of MINE.

    Parameters
    ----------
    targets : List[dict]
        Listt of target dictionaries to write.
    db : MINE
        MINE object to write targets with.
    chunk_size : int, optional
        Size of chunks to break compounds into when writing, by default 10000.
    """

    def _get_cpd_insert(cpd_dict: dict):
        output_keys = ["_id", "ID", "SMILES", "InChI_key"]
        return pymongo.InsertOne(
            {key: cpd_dict.get(key) for key in output_keys if cpd_dict.get(key) != None}
        )

    n_cpds = len(targets)
    for i, target_chunk in enumerate(utils.Chunks(targets, chunk_size)):
        if i % 20 == 0:
            print(f"Writing Targets: Chunk {i} of {int(n_cpds/chunk_size) + 1}")
        cpd_requests = [_get_cpd_insert(cpd_dict) for cpd_dict in target_chunk]
        db.target_compounds.bulk_write(cpd_requests, ordered=False)



from itertools import product
from pathlib import Path
from typing import List, Set

import numpy as np
import rdkit.rdBase as rkrb
import rdkit.RDLogger as rkl
import torch
import torch.utils.data
from keras.models import model_from_json
from rdkit.Chem import AllChem, MolFromSmiles, MolToSmiles, RemoveHs
from torch import nn
from torch.nn import functional as F

from minedatabase.filters.base_filter import Filter
from minedatabase.pickaxe import Pickaxe


logger = rkl.logger()
logger.setLevel(rkl.ERROR)
rkrb.DisableLog("rdApp.error")

device = torch.device("cpu")

CHARSET = [
    " ",
    "#",
    "(",
    ")",
    "+",
    "-",
    "/",
    "1",
    "2",
    "3",
    "4",
    "5",
    "6",
    "7",
    "8",
    "=",
    "@",
    "B",
    "C",
    "F",
    "H",
    "I",
    "N",
    "O",
    "P",
    "S",
    "[",
    "\\",
    "]",
    "c",
    "l",
    "n",
    "o",
    "r",
    "s",
]

cwd = Path(__file__).parent


class OneHotFeaturizer(object):
    def __init__(self, charset=CHARSET, padlength=120):
        self.charset = CHARSET
        self.pad_length = padlength

    def featurize(self, smiles):
        return np.array([self.one_hot_encode(smi) for smi in smiles])

    def one_hot_array(self, i):
        return [int(x) for x in [ix == i for ix in range(len(self.charset))]]

    def one_hot_index(self, c):
        return self.charset.index(c)

    def pad_smi(self, smi):
        return smi.ljust(self.pad_length)

    def one_hot_encode(self, smi):
        return np.array(
            [self.one_hot_array(self.one_hot_index(x)) for x in self.pad_smi(smi)]
        )

    def one_hot_decode(self, z):
        z1 = []
        for i in range(len(z)):
            s = ""
            for j in range(len(z[i])):
                oh = np.argmax(z[i][j])
                s += self.charset[oh]
            z1.append([s.strip()])
        return z1

    def decode_smiles_from_index(self, vec):
        return "".join(map(lambda x: CHARSET[x], vec)).strip()


class ReactionFeasibilityFilter(Filter):
    """A filter that filters compounds based on their reaction feasibility.

    Uses the following analysis Biotechnol J. 2021 May;16(5):e2000605. doi: 10.1002/biot.202000605

    Parameters
    ----------
    use_unpredicted : Bool
        Whether or not use to unpredicted reactions, by default False
    use_wildcards : Bool
        Whether or not to allow wildcard reactions through, by default True
    generation_list : list
        Generations to apply filter -- empty list filters all, by default empty list
    last_generation_only : bool
        Whether or not to only filter last generation, by default False

    Attributes
    ----------
    use_unpredicted : bool
        Whether or not use to unpredicted reactions, by default False
    use_wildcards : Bool
        Whether or not to allow wildcard reactions through, by default True
    generation_list : list
        Generations to apply filter -- empty list filters all, by default empty list
    last_generation_only : bool
        Whether or not to only filter last generation, by default False
    """

    def __init__(
        self,
        use_unpredicted: bool = False,
        use_wildcards: bool = True,
        generation_list: List[int] = [],
        last_generation_only: bool = False,
    ) -> None:
        # Import modules that are only required for the RFC
        self._filter_name = "Reaction Feasibility Filter"
        self.use_unpredicted = use_unpredicted
        self.generation_list = generation_list
        self.last_generation_only = last_generation_only

        # Threshold for feasibility
        self._feas_threshold = 0.32

    @property
    def filter_name(self) -> str:
        return self._filter_name

    def _pre_print(self) -> None:
        """Print before filtering."""
        print(f"Filter out reactions for feasibility.")

    def _post_print(
        self, pickaxe: Pickaxe, n_total: int, n_filtered: int, time_sample: float
    ) -> None:
        """Print after filtering."""
        print(
            (
                f"{n_filtered} of {n_total} "
                "compounds selected after feasibility filtering in time_sample."
            )
        )

    def _get_inputs(self, rxn_list, pickaxe):
        """rxn_list will be pickaxe eventually"""
        # Get reactions information
        def get_cpd_smiles(cpd_id):
            return pickaxe.compounds[cpd_id]["SMILES"]

        reactions_info = {}
        for rxn_id in rxn_list:
            rxn = pickaxe.reactions[rxn_id]
            reactants = [
                get_cpd_smiles(v[1]) for v in rxn["Reactants"] if v[1].startswith("C")
            ]
            products = [
                get_cpd_smiles(v[1]) for v in rxn["Products"] if v[1].startswith("C")
            ]

            pairs = product(reactants, products)
            reactions_info[rxn["_id"]] = list(pairs)

        # Process this information
        input_info = {}
        input_fails = {}
        for rxn_id, reaction_pairs in reactions_info.items():
            if not reaction_pairs:
                continue

            for i, (reactant_smiles, product_smiles) in enumerate(reaction_pairs):
                if len(reactant_smiles) <= 120:
                    if len(product_smiles) <= 120:
                        mol1 = MolFromSmiles(reactant_smiles)
                        mol2 = MolFromSmiles(product_smiles)
                        mol1 = RemoveHs(mol1)
                        mol2 = RemoveHs(mol2)
                        reactant_smiles = MolToSmiles(mol1)
                        product_smiles = MolToSmiles(mol2)

                        # TODO what does this fix? from original code
                        if "M" in reactant_smiles or "M" in product_smiles:
                            input_fails[rxn_id + "_" + str(i)] = None
                        else:
                            input_info[rxn_id + "_" + str(i)] = [
                                reactant_smiles,
                                product_smiles,
                            ]
                    else:
                        input_fails[rxn_id + "_" + str(i)] = None
                else:
                    input_fails[rxn_id + "_" + str(i)] = None

        return input_info, input_fails

    def _choose_items_to_filter(self, pickaxe: Pickaxe, processes: int = 1) -> Set[str]:
        """
        Check the compounds against the MW constraints and return
        compounds to filter.
        """

        def get_cpd_smiles(cpd_id):
            return pickaxe.compounds[cpd_id]["SMILES"]

        cpds_remove_set = set()
        rxns_remove_set = set()

        # TODO put these statements together
        # No reactions to filter for
        if len(pickaxe.reactions) == 0:
            print("No reactions to calculate feasibilty for.")
            return cpds_remove_set, rxns_remove_set

        if self.last_generation_only and pickaxe.generation != self.generation:
            print("Not filtering for this generation using feasibility.")
            return cpds_remove_set, rxns_remove_set

        if self.generation_list and (self.generation - 1) not in self.generation_list:
            print("Not filtering for this generation using feasibility.")
            return cpds_remove_set, rxns_remove_set

        reactions_to_check = []
        for cpd in pickaxe.compounds.values():
            # Compounds are in generation and correct type
            if cpd["Generation"] == pickaxe.generation and cpd["Type"] not in [
                "Coreactant",
                "Target Compound",
            ]:
                reactions_to_check.extend(cpd["Product_of"])

        reactions_to_check = set(reactions_to_check)

        rfc_input, rfc_fails = self._get_inputs(reactions_to_check, pickaxe)

        feasibility_dict = {}
        rfc_predictions = _get_feasibility(rfc_input, self._feas_threshold)

        for rxn_id, feas_scores in rfc_predictions.items():
            rxn_id = rxn_id.split("_")[0]
            # Set to true if any are feasible, set to false if doesn't
            if feas_scores[2] == "feasible":
                feasibility_dict[rxn_id] = True
            elif rxn_id not in feasibility_dict:
                feasibility_dict[rxn_id] = False

        # Assign values for unpredicted
        for rxn_id in reactions_to_check:
            rxn = pickaxe.reactions[rxn_id]
            use_rxn = False
            if rxn_id not in feasibility_dict:
                use_rxn = self.use_unpredicted

                # Check for asterisks
                reactants = [
                    get_cpd_smiles(v[1])
                    for v in rxn["Reactants"]
                    if v[1].startswith("C")
                ]
                products = [
                    get_cpd_smiles(v[1])
                    for v in rxn["Products"]
                    if v[1].startswith("C")
                ]

                compounds = reactants.extend(products)
                if any([("*" in cpd) for cpd in compounds]) and self.use_wildcards:
                    use_rxn = True

            feasibility_dict[rxn_id] = use_rxn

        rxns_remove_set = set(
            rxn_id for rxn_id, keep in feasibility_dict.items() if not keep
        )

        return cpds_remove_set, rxns_remove_set


def _get_feasibility(input_info, feas_threshold=0.32):
    # inputs
    # Load Model

    weight_file = cwd.parent / "data/feasibility/final_model.h5"
    vae_model_file = cwd.parent / "data/feasibility/vae_model.pth"
    model_file = cwd.parent / "data/feasibility/final_model.json"

    json_file = open(model_file, "r")
    loaded_model_json = json_file.read()
    json_file.close()

    loaded_model = model_from_json(loaded_model_json)
    loaded_model.load_weights(weight_file)

    vae_model = MolecularVAE()
    vae_model.load_state_dict(torch.load(vae_model_file, map_location=device))
    vae_model.to(device)
    vae_model.eval()

    feature_info = _calculate_features(input_info, vae_model)
    results = _predict_reaction_feasibility(feature_info, loaded_model, feas_threshold)

    return results


def _predict_reaction_feasibility(feature_info, model, feas_threshold):
    results = {}
    for each_label in feature_info:
        if all([type(a) == type(None) for a in feature_info[each_label]]):
            continue

        X1 = feature_info[each_label][0]
        X2 = feature_info[each_label][1]

        val_list = []
        for i in range(10):
            pred_result = model.predict([X1, X2])
            val = pred_result[0][0]
            val_list.append(val)
        mean_val = np.mean(val_list)
        std_val = np.std(val_list)
        final_val = mean_val - (std_val * 0.5)
        if final_val >= feas_threshold:
            feasibility = "feasible"
        else:
            feasibility = "infeasible"

        results[each_label] = [mean_val, std_val, feasibility]

    return results


def _calculate_features(input_info, vae_model):
    r = 2
    feature_info = {}
    for each_label in input_info:

        reactant_smi = input_info[each_label][0]
        product_smi = input_info[each_label][1]

        if "9" in reactant_smi or "9" in product_smi:
            # Does do large rings
            feature_info[each_label] = [None, None]
            continue

        z1 = _calc_z(vae_model, reactant_smi)
        z1 = z1[0]

        z2 = _calc_z(vae_model, product_smi)
        z2 = z2[0]

        reactant_mol = MolFromSmiles(reactant_smi)
        fp = AllChem.GetMorganFingerprintAsBitVect(reactant_mol, radius=r, nBits=1024)
        bits = fp.ToBitString()
        reactant_feature = []
        for f in z1:
            reactant_feature.append(float(f))
        for f in bits:
            reactant_feature.append(int(f))

        product_mol = MolFromSmiles(product_smi)
        fp = AllChem.GetMorganFingerprintAsBitVect(product_mol, radius=r, nBits=1024)
        bits = fp.ToBitString()
        product_feature = []
        for f in z2:
            product_feature.append(float(f))
        for f in bits:
            product_feature.append(int(f))

        reactant_feature = [reactant_feature]
        product_feature = [product_feature]
        reactant_feature = np.asarray(reactant_feature)
        product_feature = np.asarray(product_feature)

        feature_info[each_label] = [reactant_feature, product_feature]
    return feature_info


def _calc_z(model, smi):
    start = smi
    start = start.ljust(120)
    oh = OneHotFeaturizer()
    start_vec = torch.from_numpy(oh.featurize([start]).astype(np.float32)).to(device)
    z = model(start_vec)[1].cpu().detach().numpy()
    return z


class MolecularVAE(nn.Module):
    def __init__(self):
        super(MolecularVAE, self).__init__()

        self.conv1d1 = nn.Conv1d(120, 9, kernel_size=9)
        self.conv1d2 = nn.Conv1d(9, 9, kernel_size=9)
        self.conv1d3 = nn.Conv1d(9, 10, kernel_size=11)
        self.fc0 = nn.Linear(90, 435)
        self.fc11 = nn.Linear(435, 292)
        self.fc12 = nn.Linear(435, 292)

        self.fc2 = nn.Linear(292, 292)
        self.gru = nn.GRU(292, 501, 3, batch_first=True)
        self.fc3 = nn.Linear(501, 35)

    def encode(self, x):
        h = F.relu(self.conv1d1(x))
        h = F.relu(self.conv1d2(h))
        h = F.relu(self.conv1d3(h))
        h = h.view(h.size(0), -1)
        h = F.selu(self.fc0(h))
        return self.fc11(h), self.fc12(h)

    def reparametrize(self, mu, logvar):
        if self.training:
            std = torch.exp(0.5 * logvar)
            eps = 1e-2 * torch.randn_like(std)
            w = eps.mul(std).add_(mu)
            return w
        else:
            return mu

    def decode(self, z):
        z = F.selu(self.fc2(z))
        z = z.view(z.size(0), 1, z.size(-1)).repeat(1, 120, 1)
        out, h = self.gru(z)
        out_reshape = out.contiguous().view(-1, out.size(-1))
        y0 = F.softmax(self.fc3(out_reshape), dim=1)
        y = y0.contiguous().view(out.size(0), -1, y0.size(-1))
        return y

    def forward(self, x):
        mu, logvar = self.encode(x)
        z = self.reparametrize(mu, logvar)
        return self.decode(z), mu, logvar

    def sample(self, x, decode_attempts=1000, noise_norm=1.0):
        target_atoms = ["C", "N", "O", "S", "F", "Cl", "Br", "H"]

        oh = OneHotFeaturizer()
        mu, logvar = self.encode(x)
        z = self.reparametrize(mu, logvar)
        z2 = z.cpu().detach().numpy()
        smiles_results = {}
        for i in range(decode_attempts):
            Z = self.perturb_z(z2, noise_norm, False)
            z3 = torch.from_numpy(Z.astype(np.float32)).to("cuda")
            X = self.decode(z3)
            X = X.cpu().detach().numpy()
            y = np.argmax(X, axis=2)
            smiles = oh.decode_smiles_from_index(y[0])
            mol = MolFromSmiles(smiles)
            if mol != None:
                smiles_results[smiles] = mol

        for i in range(decode_attempts):
            Z = self.perturb_z(z2, noise_norm, True)
            z3 = torch.from_numpy(Z.astype(np.float32)).to("cuda")
            X = self.decode(z3)
            X = X.cpu().detach().numpy()
            y = np.argmax(X, axis=2)
            smiles = oh.decode_smiles_from_index(y[0])
            mol = MolFromSmiles(smiles)
            if mol != None:
                smiles_results[smiles] = mol

        for smiles in smiles_results:
            flag = True
            each_tmp_mol = smiles_results[smiles]
            for each_atom in each_tmp_mol.GetAtoms():
                if each_atom.GetSymbol() not in target_atoms:
                    flag = False
            if flag:
                print(smiles)
        return

    def perturb_z(self, z, noise_norm, constant_norm=False):
        if noise_norm > 0.0:
            noise_vec = np.random.normal(0, 1, size=z.shape)
            noise_vec = noise_vec / np.linalg.norm(noise_vec)
            if constant_norm:
                return z + (noise_norm * noise_vec)
            else:
                noise_amp = np.random.uniform(0, noise_norm, size=(z.shape[0], 1))
                return z + (noise_amp * noise_vec)
        else:
            return z


if __name__ == "__main__":
    pass



from typing import Set

import rdkit.rdBase as rkrb
import rdkit.RDLogger as rkl
from equilibrator_api import Q_

from minedatabase.filters.base_filter import Filter
from minedatabase.pickaxe import Pickaxe
from minedatabase.thermodynamics import Thermodynamics


logger = rkl.logger()
logger.setLevel(rkl.ERROR)
rkrb.DisableLog("rdApp.error")


class ThermoFilter(Filter):
    """A filter that removes reactions and compounds with bad ∆Gr

    This filter allows for the specification of a pH, Ionic strength, pMg and using
    these to calculate ∆Gr. Reeactions are then filtered out based on ∆Gr.

    Parameters
    ----------
    eq_uri : str
        URI of ccache. Can be an sqlite or postgres URI. If no uri is given
        the default is the system default eQuilibrator uri.
    dg_max : float
        Maximum ∆Gr in kJ/mol, by default 0.
    pH : float
        pH of the expansion, by default 7
    ionic_strength : float
        ionic strength of the expansion, by default 0
    pMg : float
        pMg of the expansion, by default 3
    generation_list : list
        Generations to apply filter -- empty list filters all, by default empty list
    last_generation_only : bool
        Whether or not to only filter last generation, by default False

    Attributes
    ----------
    dg_max : Q_
    pH : Q_
        pH of the expansion
    ionic_strength : Q_
        ionic strength of the expansion
    pMg : Q_
        pMg of the expansion
    generation_list : list
        Generations to apply filter -- empty list filters all, by default empty list
    last_generation_only : bool
        Whether or not to only filter last generation, by default False
    """

    def __init__(
        self,
        eq_uri=None,
        dg_max=0,
        p_h=7,
        ionic_strength=0,
        p_mg=3,
        physiological=False,
        generation_list=[],
        last_generation_only=False,
    ) -> None:
        self._filter_name = "Thermodynamic Filter"
        self.dg_max = Q_(f"{dg_max}kJ/mol")
        self.p_h = Q_(f"{p_h}")
        self.ionic_strength = Q_(f"{ionic_strength}M")
        self.p_mg = Q_(f"{p_mg}")
        self.physiological = physiological
        self.generation_list = generation_list
        self.last_generation_only = last_generation_only

        self.thermo = Thermodynamics()
        if not eq_uri:
            eq_uri = ""
        if "post" in eq_uri:
            self.thermo.load_thermo_from_postgres(eq_uri)
        elif "sql" in eq_uri:
            self.thermo.load_thermo_from_sqlite(eq_uri)
        else:
            self.thermo.load_thermo_from_sqlite()

    @property
    def filter_name(self) -> str:
        return self._filter_name

    def _pre_print(self) -> None:
        """Print before filtering."""
        print(f"Filter out reactions with ∆Gr < {self.dg_max}")

    def _post_print(
        self, pickaxe: Pickaxe, n_total: int, n_filtered: int, time_sample: float
    ) -> None:
        """Print after filtering."""
        print(
            (
                f"{n_filtered} of {n_total} "
                "compounds selected after thermodynamic filtering in time_sample."
            )
        )

    def _choose_items_to_filter(self, pickaxe: Pickaxe, processes: int = 1) -> Set[str]:
        """
        Check the compounds against the MW constraints and return
        compounds to filter.
        """
        cpds_remove_set = set()
        rxns_remove_set = set()

        # TODO put these statements together
        # No reactions to filter for
        if len(pickaxe.reactions) == 0:
            print("No reactions to calculate ∆Gr for.")
            return cpds_remove_set, rxns_remove_set

        if self.last_generation_only and pickaxe.generation != self.generation:
            print("Not filtering for this generation using thermodynamics.")
            return cpds_remove_set, rxns_remove_set

        if self.generation_list and (self.generation - 1) not in self.generation_list:
            print("Not filtering for this generation using thermodynamics.")
            return cpds_remove_set, rxns_remove_set

        print(
            f"Filtering Generation {pickaxe.generation} "
            f"with ∆G <= {self.dg_max} at pH={self.p_h}, "
            f"I={self.ionic_strength}, pMg={self.p_mg}"
        )

        reactions_to_check = []
        for cpd in pickaxe.compounds.values():
            # Compounds are in generation and correct type
            if cpd["Generation"] == pickaxe.generation and cpd["Type"] not in [
                "Coreactant",
                "Target Compound",
            ]:
                reactions_to_check.extend(cpd["Product_of"])

        reactions_to_check = set(reactions_to_check)

        for rxn_id in reactions_to_check:
            if self.physiological:
                rxn_dg = self.thermo.physiological_dg_prime_from_rid(
                    r_id=rxn_id, pickaxe=pickaxe
                )
            else:
                rxn_dg = self.thermo.dg_prime_from_rid(
                    r_id=rxn_id,
                    pickaxe=pickaxe,
                    p_h=Q_(f"{self.p_h}"),
                    ionic_strength=Q_(f"{self.ionic_strength}"),
                    p_mg=Q_(f"{self.p_mg}"),
                )
            if rxn_dg >= self.dg_max:
                rxns_remove_set.add(rxn_id)

        return cpds_remove_set, rxns_remove_set



"""Utils.py: contains basic functions reused in various contexts in other
modules"""
import collections
import csv
import hashlib
import itertools
import json
import re
import sys
from collections import deque
from collections.abc import Iterable, Iterator, Mapping, Set
from itertools import chain, islice
from numbers import Number
from os import path
from typing import List, Tuple, Union

import rdkit
from rdkit import Chem
from rdkit.Chem import AllChem


StoichTuple = collections.namedtuple("StoichTuple", "stoich,c_id")


class Chunks(Iterator):
    """A class to chunk an iterator up into defined sizes."""

    def __init__(
        self, it: Iterable, chunk_size: int = 1, return_list: bool = False
    ) -> None:
        self._it = iter(it)
        self._chunk_size = chunk_size
        self._return_list = return_list

    def __iter__(self) -> Iterator:
        return self

    def __next__(self):
        return self.next()

    def next(self) -> Union[List[chain], chain]:
        """Returns the next chunk from the iterable.
        This method is not thread-safe.

        Returns
        -------
        next_slice : Union[List[chain], chain]
            Next chunk.
        """

        def peek(iterable: Iterable) -> chain:
            """peek at first element of iterable to determine if it is empty."""
            try:
                first = next(iterable)
            except StopIteration:
                return None
            return chain([first], iterable)

        next_slice = islice(self._it, self._chunk_size)
        next_slice = peek(next_slice)

        if next_slice:
            if self._return_list:
                return list(next_slice)
            else:
                return next_slice
        else:
            raise StopIteration


def file_to_dict_list(filepath: str) -> list:
    """Accept a path to a CSV, TSV or JSON file and return a dictionary list.

    Parameters
    ----------
    filepath : str
        File to load into a dictionary list.

    Returns
    -------
    list
        Dictionary list.
    """
    filepath = str(filepath)

    if ".tsv" in filepath:
        reader = csv.DictReader(open(filepath), dialect="excel-tab")
    elif ".csv" in filepath:
        reader = csv.DictReader(open(filepath))
    elif ".json" in filepath:
        reader = json.load(open(filepath))
    else:
        raise ValueError("Unrecognized input file type")
    return list(reader)


def get_fp(smi: str) -> AllChem.RDKFingerprint:
    """Generate default RDKFingerprint.

    Parameters
    ----------
    smi : str
        SMILES of the molecule.

    Returns
    -------
    AllChem.RDKFingerprint
        Default fingerprint of the molecule.
    """
    mol = AllChem.MolFromSmiles(smi)
    fp = AllChem.RDKFingerprint(mol)
    return fp


def get_compound_hash(
    smi: str, cpd_type: str = "Predicted", inchi_blocks: int = 1
) -> Tuple[str, Union[str, None]]:
    """Create a hash string for a given compound.

    This function generates an unique identifier for a compound, ensuring a
    normalized SMILES. The compound hash is generated by sanitizing and neutralizing the
    SMILES and then generating a hash from the sha1 method in the haslib.

    The hash is prepended with a character depending on the type. Default value is "C":
        1. Coreactant: "X"
        2. Target Compound: "T"
        3. Predicted Compound: "C"

    Parameters
    ----------
    smi : str
        The SMILES of the compound.
    cpd_type : str, optional
        The Compound Type, by default 'Predicted'.

    Returns
    -------
    Tuple[str, Union[str, None]]
        Compound hash, InChI-Key.
    """

    # The ID is generated from a hash of either the InChI key (partial) or SMILES
    # The InChI key is used if the SMILES does not contain '*'
    inchi_key = None

    if "*" not in smi:
        compound = AllChem.MolFromSmiles(smi)
        inchi_key = AllChem.MolToInchiKey(compound)
        # Take the first part of the InChIKey as it contains structural
        # information only
        compound = inchi_key.rsplit("-", 3 - inchi_blocks)[0]
    else:
        compound = smi

    # Create standard length hash string using hashlib module
    chash = hashlib.sha1(compound.encode("utf-8")).hexdigest()

    # Mark cofactors with an X at the beginning, targets with a T, all else with a C
    if cpd_type == "Coreactant":
        return "X" + chash, compound
    elif cpd_type == "Target Compound":
        return "T" + chash, inchi_key
    else:
        return "C" + chash, inchi_key


def get_size(obj_0):
    """Recursively iterate to sum size of object & members."""
    ZERO_DEPTH_BASES = (str, bytes, Number, range, bytearray)
    _seen_ids = set()

    def inner(obj):
        obj_id = id(obj)
        if obj_id in _seen_ids:
            return 0
        _seen_ids.add(obj_id)
        size = sys.getsizeof(obj)
        if isinstance(obj, ZERO_DEPTH_BASES):
            pass  # bypass remaining control flow and return
        elif isinstance(obj, (tuple, list, Set, deque)):
            size += sum(inner(i) for i in obj)
        elif isinstance(obj, Mapping) or hasattr(obj, "items"):
            size += sum(inner(k) + inner(v) for k, v in getattr(obj, "items")())
        # Check for custom object instances - may subclass above too
        if hasattr(obj, "__dict__"):
            size += inner(vars(obj))
        if hasattr(obj, "__slots__"):  # can have __slots__ with __dict__
            size += sum(
                inner(getattr(obj, s)) for s in obj.__slots__ if hasattr(obj, s)
            )
        return size

    return inner(obj_0)


def convert_sets_to_lists(obj: dict) -> dict:
    """Recursively converts dictionaries that contain sets to lists.

    Parameters
    ----------
    obj : dict
        Input object to convert sets from.

    Returns
    -------
    dict
        dictionary with no sets.
    """
    if isinstance(obj, set):
        # This brings short names to the top of the list
        try:
            obj = sorted(list(obj), key=len)
        except TypeError:
            obj = list(obj)
    elif isinstance(obj, dict):
        for key in obj:
            obj[key] = convert_sets_to_lists(obj[key])
    return obj


def get_dotted_field(input_dict: dict, accessor_string: str) -> dict:
    """Gets data from a dictionary using a dotted accessor-string.

    Parameters
    ----------
    input_dict : dict
        A nested dictionary.
    accessor_string : str
        The value in the nested dict.

    Returns
    -------
    dict
        Data from the dictionary.
    """
    current_data = input_dict
    for chunk in accessor_string.split("."):
        current_data = current_data.get(chunk, {})
    return current_data


def save_dotted_field(accessor_string: str, data: dict):
    """Saves data to a dictionary using a dotted accessor-string.

    Parameters
    ----------
    accessor_string : str
        A dotted path description, e.g. "DBLinks.KEGG".
    data : dict
        The value to be stored.

    Returns
    -------
    dict
        The nested dictionary.
    """
    for chunk in accessor_string.split(".")[::-1]:
        data = {chunk: data}
    return data


def prevent_overwrite(write_path: str) -> str:
    """Prevents overwrite of existing output files by appending "_new" when
    needed.

    Parameters
    ----------
    write_path : str
        Path to write.

    Returns
    -------
    str
        Updated path to write.
    """
    write_path = str(write_path)
    while path.exists(write_path):
        split = write_path.split(".")
        # Make sure that files without an extension are still valid (otherwise,
        # split would create a list of one string which would give an index
        # error when sp[-2] is called)
        if len(split) > 1:
            split[-2] += "_new"
            write_path = ".".join(split)
        else:
            write_path += "_new"
    return write_path


# TODO: Marked for deletion
# def dict_merge(finaldict: dict, sourcedict: dict) -> None:
#     """Merges two dictionaries using sets to avoid duplication of values.

#     Parameters
#     ----------
#     finaldict : dict
#         Dict to merge into.
#     sourcedict : dict
#         Dict to merge from.
#     """
#     for key, val in sourcedict.items():
#         if (key in finaldict) and isinstance(finaldict[key], list):
#             finaldict[key] = set(finaldict[key])
#         if isinstance(val, list):
#             if key in finaldict:
#                 finaldict[key].update(val)
#             else:
#                 finaldict[key] = set(val)
#         elif isinstance(val, str):
#             if key in finaldict:
#                 finaldict[key].update(val)
#             else:
#                 finaldict[key] = set(val)
#                 finaldict[key].update(val)
#         elif isinstance(val, float):
#             if key not in finaldict:
#                 finaldict[key] = val
#         elif isinstance(val, dict):
#             if key not in finaldict:
#                 finaldict[key] = {}
#             dict_merge(finaldict[key], val)


def get_reaction_hash(
    reactants: List[StoichTuple], products: List[StoichTuple]
) -> Tuple[str, str]:
    """Hashes reactant and product lists.

    Generates a unique ID for a given reaction for use in MongoDB.

    Parameters
    ----------
    reactants : List[StoichTuple]
        List of reactants.
    products : List[StoichTuple]
        List of products.

    Returns
    -------
    Tuple[str, str]
        Reaction hash and SMILES.
    """
    # Get text reaction to be hashed
    # this is a combination of two functions
    def to_str(half_rxn):
        return [
            f"({x[0]}) {x[1]})"
            if (len(x) == 2 and not isinstance(x, str))
            else f"(1) {x}"
            for x in sorted(half_rxn)
        ]

    def get_smiles(cpds):
        cpd_tups = [
            (stoich, cpd_dict["_id"], cpd_dict["SMILES"]) for stoich, cpd_dict in cpds
        ]
        cpd_tups.sort(key=lambda x: x[1])
        smiles = []
        for cpd in cpd_tups:
            smiles.append(f"({cpd[0]}) {cpd[2]}")
        return " + ".join(smiles)

    reactant_ids = [reactant[1]["_id"] for reactant in reactants]
    product_ids = [product[1]["_id"] for product in products]
    reactant_ids.sort()
    product_ids.sort()
    text_ids_rxn = (
        " + ".join(to_str(reactant_ids)) + " => " + " + ".join(to_str(product_ids))
    )
    # Hash text reaction
    rhash = "R" + hashlib.sha256(text_ids_rxn.encode()).hexdigest()

    # smiles
    reactant_smiles = get_smiles(reactants)
    product_smiles = get_smiles(products)

    text_smiles_rxn = reactant_smiles + " => " + product_smiles

    return rhash, text_smiles_rxn


_REACTIONS = None


def neutralise_charges(
    mol: rdkit.Chem.rdchem.Mol, reactions=None
) -> rdkit.Chem.rdchem.Mol:
    """Neutralize all charges in an rdkit mol.

    Parameters
    ----------
    mol : rdkit.Chem.rdchem.Mol
        Molecule to neutralize.
    reactions : list, optional
        patterns to neutralize, by default None.

    Returns
    -------
    mol : rdkit.Chem.rdchem.Mol
        Neutralized molecule.
    """

    def _initialise_neutralisation_reactions():
        patts = (
            # Imidazoles
            ("[n+;H]", "n"),
            # Amines
            ("[N+;!H0]", "N"),
            # Carboxylic acids and alcohols
            ("[$([O-]);!$([O-][#7])]", "O"),
            # Thiols
            ("[S-;X1]", "S"),
            # Sulfonamides
            ("[$([N-;X2]S(=O)=O)]", "N"),
            # Enamines
            ("[$([N-;X2][C,N]=C)]", "N"),
            # Tetrazoles
            ("[n-]", "[nH]"),
            # Sulfoxides
            ("[$([S-]=O)]", "S"),
            # Amides
            ("[$([N-]C=O)]", "N"),
        )
        return [
            (AllChem.MolFromSmarts(x), AllChem.MolFromSmiles(y, False))
            for x, y in patts
        ]

    global _REACTIONS  # pylint: disable=global-statement
    if reactions is None:
        if _REACTIONS is None:
            _REACTIONS = _initialise_neutralisation_reactions()
        reactions = _REACTIONS
    for (reactant, product) in reactions:
        while mol.HasSubstructMatch(reactant):
            rms = AllChem.ReplaceSubstructs(mol, reactant, product)
            mol = rms[0]
    return mol


def postsanitize_smiles(smiles_list):
    """Postsanitize smiles after running SMARTS.
    :returns tautomer list of list of smiles"""
    sanitized_list = []
    tautomer_smarts = "[#7H1X3&a:1]:[#6&a:2]:[#7H0X2&a:3]>>[#7H0X2:1]:[#6:2]:[#7H1X3:3]"
    for s in smiles_list:
        temp_mol = Chem.MolFromSmiles(s, sanitize=False)
        aromatic_bonds = [
            i.GetIdx()
            for i in temp_mol.GetBonds()
            if i.GetBondType() == Chem.rdchem.BondType.AROMATIC
        ]
        for i in temp_mol.GetBonds():
            if i.GetBondType() == Chem.rdchem.BondType.UNSPECIFIED:
                i.SetBondType(Chem.rdchem.BondType.SINGLE)
        try:
            Chem.SanitizeMol(temp_mol)
            Chem.rdmolops.RemoveStereochemistry(temp_mol)
            temp_smiles = Chem.MolToSmiles(temp_mol)
        except Exception as msg:
            if "Can't kekulize mol" in str(msg):
                pyrrole_indices = [
                    i[0] for i in temp_mol.GetSubstructMatches(Chem.MolFromSmarts("n"))
                ]
                # indices to sanitize
                for s_i in pyrrole_indices:
                    temp_mol = Chem.MolFromSmiles(s, sanitize=False)
                    if temp_mol.GetAtomWithIdx(s_i).GetNumExplicitHs() == 0:
                        temp_mol.GetAtomWithIdx(s_i).SetNumExplicitHs(1)
                    elif temp_mol.GetAtomWithIdx(s_i).GetNumExplicitHs() == 1:
                        temp_mol.GetAtomWithIdx(s_i).SetNumExplicitHs(0)
                    try:
                        Chem.SanitizeMol(temp_mol)
                        processed_pyrrole_indices = [
                            i[0]
                            for i in temp_mol.GetSubstructMatches(
                                Chem.MolFromSmarts("n")
                            )
                        ]
                        processed_aromatic_bonds = [
                            i.GetIdx()
                            for i in temp_mol.GetBonds()
                            if i.GetBondType() == Chem.rdchem.BondType.AROMATIC
                        ]
                        if (
                            processed_pyrrole_indices != pyrrole_indices
                            or aromatic_bonds != processed_aromatic_bonds
                        ):
                            continue
                        Chem.rdmolops.RemoveStereochemistry(temp_mol)
                        temp_smiles = Chem.MolToSmiles(temp_mol)
                        break
                    except:
                        continue
                if "temp_smiles" not in vars():
                    Chem.rdmolops.RemoveStereochemistry(temp_mol)
                    temp_smiles = Chem.MolToSmiles(temp_mol)
                    sanitized_list.append([temp_smiles])
                    continue
            else:
                Chem.rdmolops.RemoveStereochemistry(temp_mol)
                temp_smiles = Chem.MolToSmiles(temp_mol)
                sanitized_list.append([temp_smiles])
                continue
        rxn = AllChem.ReactionFromSmarts(tautomer_smarts)
        try:
            tautomer_mols = rxn.RunReactants((Chem.MolFromSmiles(temp_smiles),))
        except:
            try:
                tautomer_mols = rxn.RunReactants(
                    (Chem.MolFromSmiles(temp_smiles, sanitize=False),)
                )
            except:
                continue
        tautomer_smiles = [Chem.MolToSmiles(m[0]) for m in tautomer_mols]
        sanitized_list.append(sorted(set(tautomer_smiles + [temp_smiles])))
    return list(itertools.product(*sanitized_list))


def get_atom_count(
    mol: rdkit.Chem.rdchem.Mol, radical_check: bool = False
) -> collections.Counter:
    """Takes a mol object and returns a counter with each element type in the set.

    Parameters
    ----------
    mol : rdkit.Chem.rdchem.Mol
        Mol object to count atoms for.
    radical_check : bool, optional
        Check for radical electrons and count if present.

    Returns
    -------
    atoms : collections.Counter
        Count of each atom type in input molecule.
    """
    atoms = collections.Counter()
    # Find all strings of the form A# in the molecular formula where A
    # is the element (e.g. C) and # is the number of atoms of that
    # element in the molecule. Pair is of form [A, #]
    for pair in re.findall(r"([A-Z][a-z]*)(\d*)", AllChem.CalcMolFormula(mol)):
        # Add # to atom count, unless there is no # (in which case
        # there is just one of that element, as ones are implicit in
        # chemical formulas)
        if pair[1]:
            atoms[pair[0]] += int(pair[1])
        else:
            atoms[pair[0]] += 1
    if radical_check:
        radical = any([atom.GetNumRadicalElectrons() for atom in mol.GetAtoms()])
        if radical:
            atoms["*"] += 1
    return atoms


def mongo_ids_to_mine_ids(mongo_ids: List[str], core_db) -> int:
    """Convert mongo ID to a MINE ID for a given compound.

    Parameters
    ----------
    mongo_id : List[str]
        List of IDs in Mongo (hashes).
    core_db : MINE
        Core database connection. Type annotation not present to avoid circular
        imports.

    Returns
    -------
    mine_id : int
        MINE ID.
    """
    mongo_to_mine = {}
    cpd_docs = core_db.compounds.find({"_id": {"$in": mongo_ids}})
    for cpd_doc in cpd_docs:
        if cpd_doc and "MINE_id" in cpd_doc:
            mine_id = cpd_doc["MINE_id"]
        else:
            mine_id = None
            print(f'Warning: {cpd_doc["MINE_id"]} not found in core DB.')
        mongo_to_mine[cpd_doc["_id"]] = mine_id
    return mongo_to_mine


# TODO: Mark for deletion.
# def _racemization(compound, max_centers=3, carbon_only=True):
#     """Enumerates all possible stereoisomers for unassigned chiral centers.

#     :param compound: A compound
#     :type compound: rdMol object
#     :param max_centers: The maximum number of unspecified stereocenters to
#         enumerate. Sterioisomers grow 2^n_centers so this cutoff prevents lag
#     :type max_centers: int
#     :param carbon_only: Only enumerate unspecified carbon centers. (other
#         centers are often not tautomeric artifacts)
#     :type carbon_only: bool
#     :return: list of stereoisomers
#     :rtype: list of rdMol objects
#     """
#     new_comps = []
#     # FindMolChiralCenters (rdkit) finds all chiral centers. We get all
#     # unassigned centers (represented by '?' in the second element
#     # of the function's return parameters).
#     unassigned_centers = [c[0] for c in AllChem.FindMolChiralCenters(
#         compound, includeUnassigned=True) if c[1] == '?']
#     # Get only unassigned centers that are carbon (atomic number of 6) if
#     # indicated
#     if carbon_only:
#         unassigned_centers = list(
#             filter(lambda x: compound.GetAtomWithIdx(x).GetAtomicNum() == 6,
#                 unassigned_centers))
#     # Return original compound if no unassigned centers exist (or if above
#     # max specified (to prevent lag))
#     if not unassigned_centers or len(unassigned_centers) > max_centers:
#         return [compound]
#     for seq in itertools.product([1, 0], repeat=len(unassigned_centers)):
#         for atomid, clockwise in zip(unassigned_centers, seq):
#             # Get both cw and ccw chiral centers for each center. Used
#             # itertools.product to get all combinations.
#             if clockwise:
#                 compound.GetAtomWithIdx(atomid).SetChiralTag(
#                     AllChem.rdchem.ChiralType.CHI_TETRAHEDRAL_CW)
#             else:
#                 compound.GetAtomWithIdx(atomid).SetChiralTag(
#                     AllChem.rdchem.ChiralType.CHI_TETRAHEDRAL_CCW)
#         # Duplicate C++ object so that we don't get multiple pointers to
#         # same object
#         new_comps.append(deepcopy(compound))
#     return new_comps



"""Fixtures used by tests from different modules go here."""

import json
from pathlib import Path

import pytest
from pymongo.errors import ServerSelectionTimeoutError

from minedatabase.databases import MINE
from minedatabase.pickaxe import Pickaxe


file_path = Path(__file__)
file_dir = file_path.parent


def delete_database(name):
    """Delete a mine database."""
    mine = MINE(name)
    mine.client.drop_database(name)
    mine.client.close()


@pytest.fixture(scope="module")
def test_db():
    """Create a test MINE database for testing."""
    datafile_path = file_dir / "data/testing_db.json"
    try:
        testdb = MINE("mongotest")
        with open(datafile_path) as infile:
            jsondb = json.load(infile)
        for doc in jsondb[0]:
            if testdb.compounds.find_one({"_id": doc["_id"]}):
                testdb.compounds.replace_one({"_id": doc["_id"]}, doc)
            else:
                testdb.compounds.insert_one(doc)
        for doc in jsondb[1]:
            if testdb.reactions.find_one({"_id": doc["_id"]}):
                testdb.reactions.replace_one({"_id": doc["_id"]}, doc)
            else:
                testdb.reactions.insert_one(doc)
        for doc in jsondb[2]:
            if testdb.operators.find_one({"_id": doc["_id"]}):
                testdb.operators.replace_one({"_id": doc["_id"]}, doc)
            else:
                testdb.operators.insert_one(doc)

    except ServerSelectionTimeoutError:
        print("No Mongo DB server detected")

    yield testdb
    delete_database("mongotest")


@pytest.fixture(scope="function")
def pk():
    """Create default Pickaxe object."""
    return Pickaxe(
        coreactant_list=file_dir / "data/test_coreactants.tsv",
        rule_list=file_dir / "data/test_reaction_rules.tsv",
        explicit_h=True,
        quiet=False,
    )


# test pickaxe
@pytest.fixture
def smiles_dict():
    """Store SMILES for compounds used in test cases here."""
    smiles = {
        "ATP": "Nc1ncnc2c1ncn2[C@@H]1O[C@H](COP(=O)(O)OP(=O)(O)OP(=O)(O)O)[C"
        + "@@H](O)[C@H]1O",
        "ADP": "Nc1ncnc2c1ncn2[C@@H]1O[C@H](COP(=O)(O)OP(=O)(O)O)[C@@H](O)[C" + "@H]1O",
        "meh": "CCC(=O)C(=O)O",
        "l_ala": "C[C@H](N)C(=O)O",
        "d_ala": "C[C@@H](N)C(=O)O",
        "FADH": "Cc1cc2c(cc1C)N(CC(O)C(O)C(O)COP(=O)(O)OP(=O)(O)OCC1OC(n3cnc"
        + "4c(N)ncnc43)C(O)C1O)c1nc(O)nc(O)c1N2",
        "S-Adenosylmethionine": "C[S+](CC[C@H](N)C(=O)O)C[C@H]1O[C@@H](n2cnc"
        + "3c(N)ncnc32)[C@H](O)[C@@H]1O",
    }
    return smiles


@pytest.fixture
def coreactant_dict(smiles_dict):
    """Create tab-formatted coreactant entries."""
    coreactants = {
        "ATP": "ATP		" + smiles_dict["ATP"],
        "ADP": "ADP		" + smiles_dict["ADP"],
        "S-Adenosylmethionine": "S-Adenosylmethionine		"
        + smiles_dict["S-Adenosylmethionine"],
    }
    return coreactants


@pytest.fixture
def default_rule(pk):
    """Set default operator."""
    return pk.operators["2.7.1.a"]


@pytest.fixture
def pk_transformed(default_rule, smiles_dict, coreactant_dict):
    """Create Pickaxe object with a few predicted reactions."""
    pk_transformed = Pickaxe(explicit_h=True)
    pk_transformed._add_compound(
        "Start", smi=smiles_dict["FADH"], cpd_type="Starting Compound"
    )
    pk_transformed._load_coreactant(coreactant_dict["ATP"])
    pk_transformed._load_coreactant(coreactant_dict["ADP"])
    pk_transformed.operators["2.7.1.a"] = default_rule
    pk_transformed.transform_all()
    pk_transformed.assign_ids()
    return pk_transformed


# Queries
@pytest.fixture()
def test_molfile():
    """Mol file for glucose compound."""
    test_molfile = open((file_dir / "data/glucose.mol"), "r").read()
    return test_molfile


@pytest.fixture()
def glucose():
    """MongoDB document (.json) for glucose compound."""
    print(file_dir / "/data/glucose.json")
    with open((file_dir / "data/glucose.json")) as infile:
        glucose = json.load(infile)
    return glucose


@pytest.fixture
def glucose_id():
    """ID in MongoDB for glucose."""
    glucose_id = {"_id": "Ccffda1b2e82fcdb0e1e710cad4d5f70df7a5d74f"}
    return glucose_id


# Filters
@pytest.fixture(scope="function")
def pk_target():
    """Generate pickaxe to expand and filter."""
    new_pk = Pickaxe(
        rule_list="./tests/data/test_filters/test_filter_rules.tsv",
        coreactant_list="./tests/data/test_filters/metacyc_coreactants.tsv",
        filter_after_final_gen=False,
    )
    new_pk.load_compound_set("./tests/data/test_filters/test_filter_compounds.csv")
    new_pk.load_targets("./tests/data/test_filters/test_filter_targets.csv")

    return new_pk



"""Basic thermodynamic calculations for pickaxe."""

from typing import Union

import pint
from equilibrator_api import (
    Q_,
    ComponentContribution,
    Reaction,
    default_physiological_ionic_strength,
    default_physiological_p_h,
    default_physiological_p_mg,
    default_physiological_temperature,
)
from equilibrator_api.phased_reaction import PhasedReaction
from equilibrator_assets.compounds import Compound
from equilibrator_assets.local_compound_cache import LocalCompoundCache
from equilibrator_cache.compound_cache import CompoundCache
from pymongo import MongoClient
from sqlalchemy import create_engine

from minedatabase.pickaxe import Pickaxe


class Thermodynamics:
    """Class to calculate thermodynamics of Pickaxe runs.

    Thermodynamics allows for the calculation of:
        1) Standard ∆G' of formation
        2) Standard ∆G'o of reaction
        3) Physiological ∆G'm of reaction
        4) Adjusted ∆G' of reaction

    eQuilibrator objects can also be obtained from r_ids and c_ids.

    Parameters
    ----------
    mongo_uri: str
        URI of the mongo database.
    client: MongoClient
        Connection to Mongo.
    CC: ComponentContribution
        eQuilibrator Component Contribution object to calculate ∆G with.
    lc: LocalCompoundCache
        The local compound cache to generate eQuilibrator compounds from.
    """

    def __init__(
        self,
    ):
        # Mongo params
        self.mongo_uri = None
        self.client = None
        self._core = None

        # eQ params
        self.CC = ComponentContribution()
        self.lc = None
        self._water = None

    def load_mongo(self, mongo_uri: Union[str, None] = None):
        if mongo_uri:
            self.mongo_uri = mongo_uri
            self.client = MongoClient(mongo_uri)
        else:
            self.mongo_uri = "localhost:27017"
            self.client = MongoClient()

        self._core = self.client["core"]

    def _all_dbs_loaded(self):
        if self.client and self._core and self.lc:
            return True
        else:
            print("Load connection to Mongo and eQuilibrator local cache.")
            return False

    def _eq_loaded(self):
        if self.lc:
            return True
        else:
            print("Load eQulibrator local cache.")
            return False

    def _reset_CC(self):
        """reset CC back to defaults"""
        self.CC.p_h = default_physiological_p_h
        self.CC.p_mg = default_physiological_p_mg
        self.CC.temperature = default_physiological_temperature
        self.CC.ionic_strength = default_physiological_ionic_strength

    def load_thermo_from_postgres(
        self, postgres_uri: str = "postgresql:///eq_compounds"
    ) -> None:
        """Load a LocalCompoundCache from a postgres uri for equilibrator.

        Parameters
        ----------
        postgres_uri : str, optional
            uri of the postgres DB to use, by default "postgresql:///eq_compounds"
        """
        self.lc = LocalCompoundCache()
        self.lc.ccache = CompoundCache(create_engine(postgres_uri))

        self._water = self.lc.get_compounds("O")

    def load_thermo_from_sqlite(
        self, sqlite_filename: str = "compounds.sqlite"
    ) -> None:
        """Load a LocalCompoundCache from a sqlite file for equilibrator.

        compounds.sqlite can be generated through LocalCompoundCache's method
        generate_local_cache_from_default_zenodo

        Parameters
        ----------
        sqlite_filename: str
            filename of the sqlite file to load.
        """
        self.lc = LocalCompoundCache()
        self.lc.load_cache(sqlite_filename)

        self._water = self.lc.get_compounds("O")

    def get_eQ_compound_from_cid(
        self, c_id: str, pickaxe: Pickaxe = None, db_name: str = None
    ) -> Union[Compound, None]:
        """Get an equilibrator compound for a given c_id from the core.

        Attempts to retrieve a compound from the core or a specified db_name.

        Parameters
        ----------
        c_id : str
            compound ID for MongoDB lookup of a compound.
        pickaxe : Pickaxe
            pickaxe object to look for the compound in, by default None.
        db_name : str
            Database to look for compound in before core database, by default None.

        Returns
        -------
        equilibrator_assets.compounds.Compound
            eQuilibrator Compound
        """
        # Find locally in pickaxe
        compound_smiles = None
        if pickaxe:
            if c_id in pickaxe.compounds:
                compound_smiles = pickaxe.compounds[c_id]["SMILES"]
            else:
                return None

        # Find in mongo db
        elif self._all_dbs_loaded():
            if db_name:
                compound = self.client[db_name].compounds.find_one(
                    {"_id": c_id}, {"SMILES": 1}
                )
                if compound:
                    compound_smiles = compound["SMILES"]

            # No cpd smiles from database name
            if not compound_smiles:
                compound = self._core.compounds.find_one({"_id": c_id}, {"SMILES": 1})
                if compound:
                    compound_smiles = compound["SMILES"]

        # No compound_smiles at all
        if not compound_smiles or "*" in compound_smiles:
            return None
        else:
            eQ_compound = self.lc.get_compounds(
                compound_smiles, bypass_chemaxon=True, save_empty_compounds=True
            )
            return eQ_compound

    def standard_dg_formation_from_cid(
        self, c_id: str, pickaxe: Pickaxe = None, db_name: str = None
    ) -> Union[float, None]:
        """Get standard ∆Gfo for a compound.

        Parameters
        ----------
        c_id : str
            Compound ID to get the ∆Gf for.
        pickaxe : Pickaxe
            pickaxe object to look for the compound in, by default None.
        db_name : str
            Database to look for compound in before core database, by default None.

        Returns
        -------
        Union[float, None]
            ∆Gf'o for a compound, or None if unavailable.
        """
        eQ_cpd = self.get_eQ_compound_from_cid(c_id, pickaxe, db_name)
        if not eQ_cpd:
            return None
        dgf = self.CC.standard_dg_formation(eQ_cpd)
        dgf = dgf[0]

        return dgf

    def get_eQ_reaction_from_rid(
        self, r_id: str, pickaxe: Pickaxe = None, db_name: str = None
    ) -> Union[PhasedReaction, None]:
        """Get an eQuilibrator reaction object from an r_id.

        Parameters
        ----------
        r_id : str
            Reaction id to get object for.
        pickaxe : Pickaxe
            pickaxe object to look for the compound in, by default None.
        db_name : str
            Database to look for reaction in.

        Returns
        -------
        PhasedReaction
            eQuilibrator reactiono to calculate ∆Gr with.
        """
        if pickaxe:
            if r_id in pickaxe.reactions:
                reaction_info = pickaxe.reactions[r_id]
            else:
                return None
        elif db_name:
            mine = self.client[db_name]
            reaction_info = mine.reactions.find_one({"_id": r_id})
            if not reaction_info:
                return None
        else:
            return None

        reactants = reaction_info["Reactants"]
        products = reaction_info["Products"]

        lhs = " + ".join(f"{r[0]} {r[1]}" for r in reactants)
        rhs = " + ".join(f"{p[0]} {p[1]}" for p in products)
        reaction_string = " => ".join([lhs, rhs])

        compounds = set([r[1] for r in reactants])
        compounds.update(tuple(p[1] for p in products))

        eQ_compound_dict = {
            c_id: self.get_eQ_compound_from_cid(c_id, pickaxe, db_name)
            for c_id in compounds
        }

        if not all(eQ_compound_dict.values()):
            return None

        if "X73bc8ef21db580aefe4dbc0af17d4013961d9d17" not in compounds:
            eQ_compound_dict["water"] = self._water

        eq_reaction = Reaction.parse_formula(eQ_compound_dict.get, reaction_string)

        return eq_reaction

    def physiological_dg_prime_from_rid(
        self, r_id: str, pickaxe: Pickaxe = None, db_name: str = None
    ) -> Union[pint.Measurement, None]:
        """Calculate the ∆Gm' of a reaction.

        Parameters
        ----------
        r_id : str
            ID of the reaction to calculate.
        pickaxe : Pickaxe
            pickaxe object to look for the compound in, by default None.
        db_name : str
            MINE the reaction is found in.

        Returns
        -------
        pint.Measurement
            The calculated ∆G'm.
        """
        eQ_reaction = self.get_eQ_reaction_from_rid(r_id, pickaxe, db_name)
        if not eQ_reaction:
            return None
        dGm_prime = self.CC.physiological_dg_prime(eQ_reaction)

        return dGm_prime

    def standard_dg_prime_from_rid(
        self, r_id: str, pickaxe: Pickaxe = None, db_name: str = None
    ) -> Union[pint.Measurement, None]:
        """Calculate the ∆G'o of a reaction.

        Parameters
        ----------
        r_id : str
            ID of the reaction to calculate.
        pickaxe : Pickaxe
            pickaxe object to look for the compound in, by default None.
        db_name : str
            MINE the reaction is found in.

        Returns
        -------
        pint.Measurement
            The calculated ∆G'o.
        """
        eQ_reaction = self.get_eQ_reaction_from_rid(r_id, pickaxe, db_name)
        if not eQ_reaction:
            return None
        dG0_prime = self.CC.standard_dg_prime(eQ_reaction)

        return dG0_prime

    def dg_prime_from_rid(
        self,
        r_id: str,
        pickaxe: Pickaxe = None,
        db_name: str = None,
        p_h: Q_ = default_physiological_p_h,
        p_mg: Q_ = default_physiological_p_mg,
        ionic_strength: Q_ = default_physiological_ionic_strength,
    ) -> Union[pint.Measurement, None]:
        """Calculate the ∆G' of a reaction.

        Parameters
        ----------
        r_id : str
            ID of the reaction to calculate.
        pickaxe : Pickaxe
            pickaxe object to look for the compound in, by default None.
        db_name : str
            MINE the reaction is found in.
        p_h : Q_
            pH of system.
        p_mg: Q_
            pMg of the system.
        ionic_strength: Q_
            ionic strength of the system.

        Returns
        -------
        pint.Measurement
            The calculated ∆G'.
        """
        eQ_reaction = self.get_eQ_reaction_from_rid(r_id, pickaxe, db_name)
        if not eQ_reaction:
            return None

        self.CC.p_h = p_h
        self.CC.p_mg = p_mg
        self.CC.ionic_strength = ionic_strength

        dG_prime = self.CC.dg_prime(eQ_reaction)

        self._reset_CC()

        return dG_prime



"""Provides functionality to interact with metabolomics datasets and search
MINE databases for metabolomics hits."""
from __future__ import annotations

import math
import os
import re
import time
import xml.etree.ElementTree as ET
from typing import Callable, Generator, List, Optional, Set, Tuple

import numpy as np
import pymongo

import minedatabase
from minedatabase.databases import MINE
from minedatabase.utils import mongo_ids_to_mine_ids


MINEDB_DIR = os.path.dirname(minedatabase.__file__)


class MetabolomicsDataset:
    """A class containing all the information for a metabolomics data set."""

    # pylint: disable=redefined-outer-name
    def __init__(
        self,
        name: str,
        adducts: List[str] = None,
        known_peaks: List[Peak] = None,
        unknown_peaks: List[Peak] = None,
        native_set: Set[str] = set(),
        ppm: bool = False,
        tolerance: float = 0.001,
        halogens: bool = False,
        verbose: bool = False,
    ):
        """Metabolomics Dataset initialization.

        Parameters
        ----------
        name : str
            Name of metabolomics dataset.
        adducts : List[str], optional
            List of adduct names, e.g. ["[M-H]-", "[M+H]+"] (defaults to all)
            See minedatabase/data/adducts/{Negative/Positive} Adducts full.txt
            for a full list of possible adducts.
        known_peaks : List[Peak], optional
            List of Peak objects annotated with ID and associated data, by
            default None.
        unknown_peaks : List[Peak], optional
            List of Peak objects annotated with associated data, by default
            None.
        native_set : Set[str], optional
            Set of compound IDs native to the organism that generated the
            dataset (e.g. IDs from model), by default set().
        ppm : bool, optional
            If True, tolerance is set in parts per million, and if false
            (default), tolerance is set in Daltons.
        tolerance : float, optional
            Mass tolerance for hits, by default 0.001.
        halogens : bool, optional
            Filters out compounds containing halogens if True, by default False.
        verbose : bool, optional
            Prints more info to stdout if True, by default False.
        """
        # Load adducts
        pos_fp = os.path.join(MINEDB_DIR, "data/adducts/Positive Adducts full.txt")
        neg_fp = os.path.join(MINEDB_DIR, "data/adducts/Negative Adducts full.txt")
        all_pos_adducts = self._read_adduct_file(pos_fp)
        all_neg_adducts = self._read_adduct_file(neg_fp)

        if adducts:
            self.pos_adducts = list(filter(lambda x: x[0] in adducts, all_pos_adducts))
            self.neg_adducts = list(filter(lambda x: x[0] in adducts, all_neg_adducts))
        else:
            self.pos_adducts = all_pos_adducts
            self.neg_adducts = all_neg_adducts

        # MongoDB projection for compound search
        self.hit_projection = {
            "Formula": 1,
            "MINE_id": 1,
            "SMILES": 1,
            "Inchikey": 1,
            "Spectra.Positive": 1,
            "Spectra.Negative": 1,
            "logP": 1,
        }

        # Load peak data and initialize other attributes
        self.name = name
        self.known_peaks = known_peaks if known_peaks else []
        self.unknown_peaks = unknown_peaks if unknown_peaks else []
        self.native_set = native_set
        self.ppm = ppm
        self.tolerance = tolerance
        self.halogens = halogens
        self.verbose = verbose

        self.total_formulas = 0
        self.total_hits = 0
        self.matched_peaks = 0
        self.possible_masses = {"+": [], "-": []}
        self.possible_ranges = {"+": [], "-": []}

    def __str__(self) -> str:
        """Give string representation.

        Returns
        -------
        self.name : str
            Name of the metabolomics dataset.
        """
        return self.name

    def _read_adduct_file(self, filepath: str) -> List[Tuple]:
        """Read specified adduct file.

        Parameters
        ----------
        filepath : str
            Path to adduct file.

        Returns
        -------
        adducts : List[Tuple]
            A list of (str, float, float) tuples of form
            ('adduct name', m/z multiplier, adduct mass change).
        """
        adducts = []
        with open(filepath, "r") as infile:
            for line in infile:
                if line.startswith("#"):
                    continue
                adduct = line.strip().split("\t")
                adduct[0] = adduct[0].strip()
                adduct[1] = float(adduct[1])
                adduct[2] = float(adduct[2])
                adducts.append(tuple(adduct))
        return adducts

    def enumerate_possible_masses(self, tolerance: float) -> None:
        """Generate all possible masses from unknown peaks and list of
        adducts. Saves these mass ranges to self.possible_ranges.

        Parameters
        ----------
        tolerance : float
            Mass tolerance in Daltons.
        """
        for peak in self.unknown_peaks:
            if peak.charge == "+":
                peak_adducts = self.pos_adducts
            else:
                peak_adducts = self.neg_adducts

            masses, ranges = peak._enumerate_possible_masses(
                self, peak_adducts, tolerance
            )
            self.possible_masses[peak.charge] += masses
            self.possible_ranges[peak.charge] += ranges

        for charge in ["+", "-"]:
            self.possible_masses[charge] = np.array(set(self.possible_masses[charge]))

    def get_rt(self, peak_id: str) -> Optional[float]:
        """Return retention time for peak with given ID. If not found, returns
        None.

        Parameters
        ----------
        peak_id : str
            ID of peak as listed in dataset.

        Returns
        -------
        rt : float, optional
            Retention time of peak with given ID, None if not found.
        """
        rt = None
        for peak in self.unknown_peaks + self.known_peaks:
            if peak_id == peak.name:
                rt = peak.r_time
                break
        return rt

    def find_db_hits(
        self,
        peak: Peak,
        db: MINE,
        core_db: MINE,
        adducts: List[Tuple[str, float, float]],
    ) -> None:
        """This function searches the database for matches of a peak given
        adducts and updates the peak object with that information.

        Parameters
        ----------
        peak : Peak
            Peak object to query against MINE compound database.
        db : MINE
            MINE database to query.
        adducts : List[Tuple[str, float, float]]
            List of adducts. Each adduct contains three values in a tuple:
            (adduct name, mass multiplier, ion mass).
        """
        # find nominal mass for a given m/z for each adduct and the max and
        # min values for db
        potential_masses = [(peak.mz - adduct[2]) / adduct[1] for adduct in adducts]

        if self.ppm:
            precision = (self.tolerance / 100000.0) * potential_masses
        else:
            precision = self.tolerance * 0.001  # convert to mDa
        upper_bounds = [pm + precision for pm in potential_masses]
        lower_bounds = [pm - precision for pm in potential_masses]

        # search database for hits in the each adducts mass range that have no
        # innate charge.
        mongo_ids = []
        for i, adduct in enumerate(adducts):
            # build the query by adding the optional terms
            query_terms = [
                {"Mass": {"$gte": float(lower_bounds[i])}},
                {"Mass": {"$lte": float(upper_bounds[i])}},
                {"Charge": 0},
                {"MINES": {"$eq": db.name}},
            ]
            if adduct[0] == "[M]+":
                query_terms[2] = {"Charge": 1}
            for compound in core_db.compounds.find(
                {"$and": query_terms}, self.hit_projection
            ):
                # Filters out halogens if the flag is enabled by moving to the
                # next compound before the current compound is counted or
                # stored.
                if not self.halogens:
                    if re.search("F[^e]|Cl|Br", compound["Formula"]):
                        continue

                # update the total hits for the peak and make a note if the
                # compound is in the native_set
                peak.total_hits += 1

                compound["native_hit"] = False
                if compound["_id"] in self.native_set:
                    peak.native_hit = True
                    compound["native_hit"] = True

                peak.formulas.add(compound["Formula"])
                compound["adduct"] = adduct[0]
                compound["peak_name"] = peak.name
                mongo_ids.append(compound["_id"])
                peak.isomers.append(compound)

        if self.native_set:
            cpd_ids = [cpd["_id"] for cpd in peak.isomers]
            native_product_ids = set(self.check_product_of_native(cpd_ids, db))
        else:
            native_product_ids = set()

        # Get MINE IDs in bulk
        mongo_to_mine = mongo_ids_to_mine_ids(mongo_ids, core_db)
        for cpd in peak.isomers:
            cpd["MINE_id"] = mongo_to_mine[cpd["_id"]]
            if cpd["_id"] in native_product_ids:
                cpd["product_of_native_hit"] = True
            else:
                cpd["product_of_native_hit"] = False

    def check_product_of_native(self, cpd_ids: List[str], db: MINE) -> List[str]:
        """Filters list of compound IDs to just those associated with compounds
        produced from a native hit in the model (i.e. in native set)."""

        pipeline = [
            {"$match": {"_id": {"$in": cpd_ids}}},
            {
                "$project": {
                    "Product_of": 1,
                }
            },
            {"$unwind": "$Product_of"},
            {
                "$lookup": {
                    "from": "product_of",
                    "localField": "Product_of",
                    "foreignField": "_id",
                    "as": "product_of_doc",
                }
            },
            {"$unwind": "$product_of_doc"},
            {
                "$project": {
                    "_id": 1,
                    "producing_rxns": "$product_of_doc.Product_of",
                }
            },
            {"$unwind": "$producing_rxns"},
            {
                "$group": {
                    "_id": "$_id",
                    "producing_rxns": {"$addToSet": "$producing_rxns"},
                    "n_rxns": {"$sum": 1},
                }
            },
            {"$unwind": "$producing_rxns"},
            {
                "$lookup": {
                    "from": "reactions",
                    "localField": "producing_rxns",
                    "foreignField": "_id",
                    "as": "reaction_doc",
                }
            },
            {
                "$project": {
                    "reactants": "$reaction_doc.Reactants",
                }
            },
            {"$project": {"reaction_doc": 0, "producing_rxns": 0}},
            {"$unwind": "$reactants"},
            {"$unwind": "$reactants"},
            {"$project": {"reactant": {"$arrayElemAt": ["$reactants", 1]}}},
            {"$match": {"reactant": {"$regex": "^C.*"}}},
            {
                "$lookup": {
                    "from": "compounds",
                    "localField": "reactant",
                    "foreignField": "_id",
                    "as": "reactant_doc",
                }
            },
            {
                "$project": {
                    "_id": 1,
                    "reactant_id": {"$arrayElemAt": ["$reactant_doc._id", 0]},
                }
            },
            {"$group": {"_id": "$_id", "reactant_ids": {"$addToSet": "$reactant_id"}}},
            {
                "$match": {
                    "reactant_ids": {"$elemMatch": {"$in": list(self.native_set)}}
                }
            },
            {"$project": {"_id": 1}},
        ]

        native_product_ids = db.compounds.aggregate(pipeline)
        native_product_ids = [doc["_id"] for doc in native_product_ids]
        return native_product_ids

    def annotate_peaks(self, db: MINE, core_db: MINE) -> None:
        """This function iterates through the unknown peaks in the dataset and
        searches the database for compounds that match a peak m/z given the
        adducts permitted. Statistics on the annotated data set are printed.

        Parameters
        ----------
        db : MINE
            MINE database.
        core_db : MINE
            Core database containing spectra info.
        """
        for i, peak in enumerate(self.unknown_peaks):

            positive = (
                peak.charge == "+"
                or peak.charge == "Positive"
                or (peak.charge and isinstance(peak.charge, bool))
            )
            negative = (
                peak.charge == "-"
                or peak.charge == "Negative"
                or (not peak.charge and isinstance(peak.charge, bool))
            )

            if positive:
                self.find_db_hits(peak, db, core_db, self.pos_adducts)
            elif negative:
                self.find_db_hits(peak, db, core_db, self.neg_adducts)
            else:
                raise ValueError(
                    "Invalid compound charge specification. "
                    'Please use "+" or "Positive" for '
                    'positive ions and "-" or "Negative" for '
                    f"negative ions. (charge = {peak.charge})"
                )

            if peak.total_hits > 0:
                self.matched_peaks += 1
                self.total_hits += peak.total_hits
                self.total_formulas += len(peak.formulas)
            if self.verbose:
                pct_done = int(float(i) / float(len(self.unknown_peaks)) * 100)
                print(f"{pct_done} percent of peaks processed")


# Scoring functions appear before the Peak class because dot_product method is
# default object for Peak.score_isomers


def dot_product(x: List[tuple], y: List[tuple], epsilon: float = 0.01) -> float:
    """Calculate the dot product of two spectra, allowing for some variability
    in mass-to-charge ratios

    Parameters
    ----------
    x : List[tuple]
        First spectra m/z values.
    y : List[tuple]
        Second spectra m/z values.
    epsilon : float, optional
        Mass tolerance in Daltons, by default 0.01.

    Returns
    -------
    dot_prod : float
        Dot product of x and y.
    """
    z = 0
    n_v1 = 0
    n_v2 = 0

    for int1, int2 in _approximate_matches(x, y, epsilon):
        z += int1 * int2
        n_v1 += int1 * int1
        n_v2 += int2 * int2

    dot_prod = z / (math.sqrt(n_v1) * math.sqrt(n_v2))
    return dot_prod


def jaccard(x: List[tuple], y: List[tuple], epsilon: float = 0.01) -> float:
    """Calculate the Jaccard Index of two spectra, allowing for some
    variability in mass-to-charge ratios

    Parameters
    ----------
    x : List[tuple]
        First spectra m/z values.
    y : List[tuple]
        Second spectra m/z values.
    epsilon : float, optional
        Mass tolerance in Daltons, by default 0.01.

    Returns
    -------
    jaccard_index : float
        Jaccard Index of x and y.
    """
    intersect = 0

    for val1, val2 in _approximate_matches(x, y, epsilon):
        if val1 and val2:
            intersect += 1

    jaccard_index = intersect / float((len(x) + len(y) - intersect))
    return jaccard_index


def _approximate_matches(
    list1: List[tuple], list2: List[tuple], epsilon: float = 0.01
) -> Generator:
    """Takes two list of tuples and searches for matches of tuples first value
    within the supplied epsilon. Emits tuples with the tuples second values
    where found. if a value in one dist does not match the other list, it is
    emitted alone but with a 0 as the other value.

    Parameters
    ----------
    list1 : list
        First list of tuples.
    list2 : list
        Second list of tuples.
    epsilon : float, optional
        Maximum difference, by default 0.01.

    Yields
    -------
    Generator
        Generator that yields found matches.
    """
    list1.sort()
    list2.sort()
    list1_index = 0
    list2_index = 0

    while list1_index < len(list1) or list2_index < len(list2):
        if list1_index == len(list1):
            yield (0, list2[list2_index][1])
            list2_index += 1
            continue
        if list2_index == len(list2):
            yield (list1[list1_index][1], 0)
            list1_index += 1
            continue

        list1_element = list1[list1_index][0]
        list2_element = list2[list2_index][0]

        difference = abs(list1_element - list2_element)

        if difference < epsilon:
            yield (list1[list1_index][1], list2[list2_index][1])
            list1_index += 1
            list2_index += 1
        elif list1_element < list2_element:
            yield (list1[list1_index][1], 0)
            list1_index += 1
        elif list2_element < list1_element:
            yield (0, list2[list2_index][1])
            list2_index += 1


class Peak:
    """Peak object which contains peak metadata as well as mass, retention
    time, spectra, and any MINE database hits.

    Parameters
    ----------
    name : str
        Name or ID of the peak.
    r_time : float
        Retention time of the peak.
    mz : float
        Mass-to-charge ratio (m/z) of the peak.
    charge : str
        Charge of the peak, "+" or "-".
    inchi_key : str, optional
        InChI key of the peak, if already identified, by default None.
    ms2 : List[float], optional
        MS2 spectra m/z values for this peak, by default None.

    Attributes
    ----------
    isomers : List[Dict]
        List of compound documents in JSON (dict) format.
    formulas : Set[str]
        All the unique compound formulas from compounds found for this peak.
    total_hits : int
        Number of compound hits for this peak.
    native_hit : bool
        Whether this peak matches a compound provided in the native set.
    """

    def __init__(
        self,
        name: str,
        r_time: float,
        mz: float,
        charge: str,
        inchi_key: str = None,
        ms2: List[(float, float)] = None,
    ) -> None:
        self.name = name
        if r_time:
            self.r_time = float(r_time)
        else:
            self.r_time = None
        self.mz = float(mz)
        self.charge = charge
        self.inchi_key = inchi_key
        self.ms2peaks = ms2

        self.isomers = []
        self.formulas = set()
        self.total_hits = 0
        self.native_hit = False

    def __str__(self) -> str:
        """String representation of the peak.

        Returns
        -------
        str
            Name of the peak.
        """
        return self.name

    def __repr__(self) -> str:
        """Print representation of the peak.

        Returns
        -------
        str
            Print representation of the peak.
        """
        return (
            f"Peak {self.name}: {self.mz} m/z, {self.r_time} RT, {self.charge} mode, "
            f"Contains {len(self.ms2peaks)} MS2 peaks starting with {self.ms2peaks[:3]}..."
        )

    def _enumerate_possible_masses(
        self, met_dataset: MetabolomicsDataset, adducts: List[str], tolerance: float
    ) -> (List[float], List[Tuple[float, float]]):
        """Generate all possible masses for a given peak.

        Parameters
        ----------
        met_dataset : MetabolomicsDataset
            Instance of MetabolomicsDataset with associated adducts.
        adducts : List[str]
            List of adducts, charge should match charge of this peak.
        tolerance : float
            Mass tolernace in Daltons.

        Returns
        -------
        possible_masses : List[float]
            List of possible masses.
        possible_ranges : List[Tuple(float, float)]
            List of lower and upper bounds provided aggregate mass + tolerance.
        """
        possible_masses = []
        possible_ranges = []
        if self.charge == "+":
            adducts = met_dataset.pos_adducts
        else:
            adducts = met_dataset.neg_adducts

        for adduct in adducts:
            possible_mass = (self.mz - adduct[2]) / adduct[1]
            possible_masses.append(possible_mass)
            possible_ranges.append(
                (
                    possible_mass - tolerance,
                    possible_mass + tolerance,
                    self.name,
                    adduct[0],
                )
            )
        return possible_masses, possible_ranges

    def score_isomers(
        self,
        metric: Callable[[list, list], float] = dot_product,
        energy_level: int = 20,
        tolerance: float = 0.005,
    ) -> None:
        """Scores and sorts isomers based on mass spectra data.

        Calculates the cosign similarity score between the provided ms2 peak
        list and pre-calculated CFM-spectra and sorts the isomer list
        according to this metric.

        Parameters
        ----------
        metric : function, optional
            The scoring metric to use for the spectra. Function must accept 2
            lists of (mz, intensity) tuples and return a score, by default dot_product.
        energy_level : int, optional
            The Fragmentation energy level to use. May be 10,
            20 or 40., by default 20.
        tolerance : float, optional
            The precision to use for matching m/z in mDa, by default 0.005.

        Raises
        ------
        ValueError
            Empty ms2 peak.
        """
        if not self.ms2peaks:
            raise ValueError("The ms2 peak list is empty")
        if self.charge == "+":
            spec_key = "Positive"
        else:
            spec_key = "Negative"

        for i, hit in enumerate(self.isomers):
            if spec_key in hit["Spectra"]:
                hit_spec = hit["Spectra"][spec_key][f"{energy_level}V"]
                score = metric(self.ms2peaks, hit_spec, epsilon=tolerance)
                rounded_score = round(score * 1000)
                self.isomers[i]["Spectral_score"] = rounded_score
            else:
                self.isomers[i]["Spectral_score"] = 0
        self.isomers.sort(key=lambda x: x["Spectral_score"], reverse=True)


def get_KEGG_comps(
    db: MINE, core_db: MINE, kegg_db: pymongo.database.Database, model_ids: List[str]
) -> set:
    """Get MINE IDs from KEGG MINE database for compounds in model(s).

    Parameters
    ----------
    db : MINE
        MINE Mongo database.
    kegg_db : pymongo.database.Database
        Mongo database with annotated organism metabolomes from KEGG.
    model_ids : List[str]
        List of organism identifiers from KEGG.

    Returns
    -------
    set
        MINE IDs of compounds that are linked to a KEGG ID in at least one of
        the organisms in model_ids.
    """
    kegg_ids, _ids = set(), set()
    for model in kegg_db.models.find({"_id": {"$in": model_ids}}):
        comp_ids = model["Compounds"]
        kegg_ids = kegg_ids.union(comp_ids)
    kegg_id_list = list(kegg_ids)  # sets are not accepted as query params for pymongo
    for comp in core_db.compounds.find(
        {"$and": [{"KEGG_id": {"$in": kegg_id_list}}, {"MINES": db.name}]}
    ):
        _ids.add(comp["_id"])
    return _ids


def read_adduct_names(filepath: str) -> List[str]:
    """Read adduct names from text file at specified path into a list.

    Parameters
    ----------
    filepath : str
        Path to adduct text file.

    Returns
    -------
    adducts : list
        Names of adducts in text file.

    Notes
    -----
    Not used in this codebase but used by MINE-Server to validate adduct input.
    """

    with open(filepath) as infile:
        adducts = [line.split(" \t")[0] for line in infile if not line[0] == "#"]

    return adducts


def read_mgf(input_string: str, charge: bool, ms2_delim="\t") -> List[Peak]:
    """Parse mgf metabolomics data file.

    Parameters
    ----------
    input_string : str
        Metabolomics input data file.
    charge : bool
        True if positive, False if negative.
    ms2_delim : str
        Delimiter for whitespace between intensity and m/z value. Usually tab
        but can also be a space in some MGF files. Tab by default.

    Returns
    -------
    peaks : List[Peak]
        A list of Peak objects.
    """
    peaks = []
    ms2 = []
    r_time = None
    for line in input_string.split("\n"):
        sl = line.strip(" \r\n").split("=")
        if sl[0] == "PEPMASS":
            if len(sl) > 1:
                mass = sl[1]
            else:
                mass = None
        elif sl[0] == "TITLE":
            if len(sl) > 1:
                name = sl[1]
            else:
                name = ""
        elif sl[0] == "RTINSECONDS":
            r_time = sl[1]
        elif sl[0] == "END IONS":
            peaks.append(Peak(name, r_time, mass, charge, "False", ms2=ms2))
            ms2 = []
        else:
            try:
                mz, i = sl[0].split(ms2_delim)
                ms2.append((float(mz), float(i)))
            except ValueError:
                continue
    return peaks


def read_msp(input_string: str, charge: bool) -> List[Peak]:
    """Parse msp metabolomics data file.

    Parameters
    ----------
    input_string : str
        Metabolomics input data file.
    charge : bool
        True if positive, False if negative.

    Returns
    -------
    peaks : List[Peak]
        A list of Peak objects.
    """
    peaks = []
    for spec in input_string.strip().split("\n\n"):
        ms2 = []
        inchikey = "False"
        r_time = 0
        name = "N/A"
        for line in spec.split("\n"):
            sl = line.split(": ")
            sl[0] = sl[0].replace(" ", "").replace("/", "").upper()
            if sl[0] == "PRECURSORMZ":
                mass = sl[1]
            elif sl[0] == "NAME":
                name = sl[1]
            elif sl[0] == "RETENTIONTIME":
                r_time = sl[1]
            elif sl[0] == "INCHIKEY":
                inchikey = sl[1]
            elif line and line[0].isdigit():
                try:
                    row = re.split("[\t ]", line)
                    ms2.append((float(row[0]), float(row[1])))
                except ValueError:
                    continue
        peaks.append(Peak(name, r_time, mass, charge, inchikey, ms2=ms2))
    return peaks


def read_mzxml(input_string: str, charge: bool) -> List[Peak]:
    """Parse mzXML metabolomics data file.

    Parameters
    ----------
    input_string : str
        Metabolomics input data file.
    charge : bool
        True if positive, False if negative.

    Returns
    -------
    List[Peak]
        A list of Peak objects.
    """
    peaks = []
    root = ET.fromstring(input_string)
    prefix = root.tag.strip("mzXML")

    for scan in root.findall(f".//{prefix}scan"):
        # somewhat counter intuitively we will get the peak info from the
        # second fragments precursor info.
        if scan.attrib["msLevel"] == "2":
            precursor = scan.find(f"./{prefix}precursorMz")
            mz = precursor.text
            r_time = scan.attrib["retentionTime"][2:-1]
            name = f"{mz} @ {r_time}"
            charge = scan.attrib["polarity"]
            peaks.append(Peak(name, r_time, mz, charge, "False"))

    return peaks


class Struct:
    """convert key-value pairs into object-attribute pairs."""

    def __init__(self, **entries):
        self.__dict__.update(entries)


def ms_adduct_search(
    db: MINE,
    core_db: MINE,
    keggdb: pymongo.database.Database,
    text: str,
    text_type: str,
    ms_params,
) -> List:
    """Search for compound-adducts matching precursor mass.

    Parameters
    ----------
    db : MINE
        Contains compound documents to search.
    core_db : MINE
        Contains extra info (including spectra) for compounds in db.
    keggdb : pymongo.database.Database
        Contains models with associated compound documents.
    text : str
        Text as in metabolomics datafile for specific peak.
    text_type : str
        Type of metabolomics datafile (mgf, mzXML, and msp are supported). If
        text, assumes m/z values are separated by newlines (and set text_type
        to "form").
    ms_params : dict
        Specifies search settings, using the following key-value pairs:
        ------------------------
        Required Key-Value Pairs
        ------------------------
        "tolerance": float specifying tolerance for m/z, in mDa by default.
            Can specify in ppm if "ppm" key's value is set to True.
        "charge": bool ('+' for positive, '-' for negative).
        ------------------------
        Optional Key-Value Pairs
        ------------------------
        "adducts": list of adducts to use. If not specified, uses all adducts.
        "models": List of model _ids. If supplied, score compounds higher if
            present in model. ["eco"] by default (E. coli).
        "ppm": bool specifying whether "tolerance" is in mDa or ppm. Default
            value for ppm is False (so tolerance is in mDa by default).
        "kovats": length 2 tuple specifying min and max kovats retention index
            to filter compounds (e.g. (500, 1000)).
        "logp": length 2 tuple specifying min and max logp to filter compounds
            (e.g. (-1, 2)).
        "halogens": bool specifying whether to filter out compounds containing
            F, Cl, or Br. Filtered out if set to True. False by default.

    Returns
    -------
    ms_adduct_output : list
        Compound JSON documents matching ms adduct query.
    """
    print(
        f"<MS Adduct Search: TextType={text_type}, Text={text}, Parameters={ms_params}>"
    )
    name = text_type + time.strftime("_%d-%m-%Y_%H:%M:%S", time.localtime())

    if isinstance(ms_params, dict):
        ms_params = Struct(**ms_params)

    dataset = MetabolomicsDataset(
        name,
        adducts=ms_params.adducts,
        ppm=ms_params.ppm,
        tolerance=ms_params.tolerance,
        halogens=ms_params.halogens,
        verbose=ms_params.verbose,
    )
    ms_adduct_output = []

    if text_type == "form":
        for mz in text.split("\n"):
            dataset.unknown_peaks.append(
                Peak(mz, 0, float(mz), ms_params.charge, "False")
            )
    elif text_type == "mgf":
        dataset.unknown_peaks = read_mgf(text, ms_params.charge)
    elif text_type == "mzXML" or text_type == "mzxml":
        dataset.unknown_peaks = read_mzxml(text, ms_params.charge)
    elif text_type == "msp":
        dataset.unknown_peaks = read_msp(text, ms_params.charge)
    else:
        raise IOError(f"{text_type} files not supported")

    if ms_params.models:
        dataset.native_set = get_KEGG_comps(db, core_db, keggdb, ms_params.models)
    else:
        dataset.native_set = set()

    dataset.annotate_peaks(db, core_db)

    if ms_params.logp:
        min_logp, max_logp = ms_params.logp
    else:
        min_logp, max_logp = (-1000, 1000)

    for peak in dataset.unknown_peaks:
        for hit in peak.isomers:
            if min_logp < hit["logP"] < max_logp:
                ms_adduct_output.append(hit)

    if ms_params.models:
        ms_adduct_output = score_compounds(
            ms_adduct_output,
            model_id=ms_params.models[0],
            mine_db=db,
            core_db=core_db,
            kegg_db=keggdb,
            parent_frac=0.75,
            reaction_frac=0.25,
            get_native=True,
        )

    return ms_adduct_output


def ms2_search(
    db: MINE,
    core_db: MINE,
    keggdb: pymongo.database.Database,
    text: str,
    text_type: str,
    ms_params,
) -> List:
    """Search for compounds matching MS2 spectra.

    Parameters
    ----------
    db : MINE
        Contains compound documents to search.
    core_db : MINE
        Contains extra info (including spectra) for compounds in db.
    keggdb : pymongo.database.Database
        Contains models with associated compound documents.
    text : str
        Text as in metabolomics datafile for specific peak.
    text_type : str
        Type of metabolomics datafile (mgf, mzXML, and msp are supported). If
        text, assumes m/z values are separated by newlines (and set text_type
        to "form").
    ms_params : dict
        Specifies search settings, using the following key-value pairs:
        ------------------------
        Required Key-Value Pairs
        ------------------------
        "tolerance": float specifying tolerance for m/z, in mDa by default.
            Can specify in ppm if "ppm" key's value is set to True.
        "charge": bool (1 for positive, 0 for negative).
        "energy_level": int specifying fragmentation energy level to use. May
            be 10, 20, or 40.
        "scoring_function": str describing which scoring function to use. Can
            be either "jaccard" or "dot product".
        ------------------------
        Optional Key-Value Pairs
        ------------------------
        "adducts": list of adducts to use. If not specified, uses all adducts.
        "models": List of model _ids. If supplied, score compounds higher if
            present in model.
        "ppm": bool specifying whether "tolerance" is in mDa or ppm. Default
            value for ppm is False (so tolerance is in mDa by default).
        "kovats": length 2 tuple specifying min and max kovats retention index
            to filter compounds (e.g. (500, 1000)).
        "logp": length 2 tuple specifying min and max logp to filter compounds
            (e.g. (-1, 2)).
        "halogens": bool specifying whether to filter out compounds containing
            F, Cl, or Br. Filtered out if set to True. False by default.

    Returns
    -------
    ms_adduct_output : list
        Compound JSON documents matching ms2 search query.
    """
    print(
        f"<MS2 Search: TextType={text_type}, Parameters={ms_params}, Spectra={repr(text)}>"
    )
    name = text_type + time.strftime("_%d-%m-%Y_%H:%M:%S", time.localtime())

    if isinstance(ms_params, dict):
        ms_params = Struct(**ms_params)

    dataset = MetabolomicsDataset(
        name,
        adducts=ms_params.adducts,
        ppm=ms_params.ppm,
        tolerance=ms_params.tolerance,
        halogens=ms_params.halogens,
        verbose=ms_params.verbose,
    )
    ms_adduct_output = []

    if text_type == "form":
        split_form = [x.split() for x in text.strip().split("\n")]
        ms2_data = [(float(mz), float(i)) for mz, i in split_form[1:]]
        peak = Peak(
            split_form[0][0],
            0,
            float(split_form[0][0]),
            ms_params.charge,
            "False",
            ms2=ms2_data,
        )
        dataset.unknown_peaks.append(peak)
    elif text_type == "mgf":
        dataset.unknown_peaks = read_mgf(text, ms_params.charge)
    elif text_type == "mzXML":
        dataset.unknown_peaks = read_mzxml(text, ms_params.charge)
    elif text_type == "msp":
        dataset.unknown_peaks = read_msp(text, ms_params.charge)
    else:
        raise IOError(f"{text_type} files not supported")

    if not ms_params.models:
        ms_params.models = ["eco"]

    if ms_params.models:
        dataset.native_set = get_KEGG_comps(db, core_db, keggdb, ms_params.models)

    dataset.annotate_peaks(db, core_db)

    if ms_params.logp:
        min_logp, max_logp = ms_params.logp
    else:
        min_logp, max_logp = (-1000, 1000)

    for peak in dataset.unknown_peaks:

        if ms_params.scoring_function == "jaccard":
            if not ms_params.ppm:
                peak.score_isomers(
                    metric=jaccard,
                    energy_level=ms_params.energy_level,
                    tolerance=float(ms_params.tolerance) / 1000,
                )
            else:
                peak.score_isomers(metric=jaccard, energy_level=ms_params.energy_level)
        elif ms_params.scoring_function == "dot product":
            if not ms_params.ppm:
                peak.score_isomers(
                    metric=dot_product,
                    energy_level=ms_params.energy_level,
                    tolerance=float(ms_params.tolerance) / 1000,
                )
            else:
                peak.score_isomers(
                    metric=dot_product, energy_level=ms_params.energy_level
                )
        else:
            raise ValueError(
                'ms_params["scoring_function"] must be either '
                '"jaccard" or "dot product".'
            )

        for hit in peak.isomers:
            if min_logp < hit["logP"] < max_logp:
                ms_adduct_output.append(hit)

        if ms_params.models:
            ms_adduct_output = score_compounds(
                ms_adduct_output,
                model_id=ms_params.models[0],
                mine_db=db,
                core_db=core_db,
                kegg_db=keggdb,
                parent_frac=0.75,
                reaction_frac=0.25,
                get_native=True,
            )

    return ms_adduct_output


def spectra_download(db: MINE, mongo_id: str = None) -> str:
    """Download one or more spectra for compounds matching a given query.

    Parameters
    ----------
    db : MINE
        Contains compound documents to search.
    mongo_query : str, optional (default: None)
        A valid Mongo query as a literal string. If None, all compound spectra
        are returned.
    parent_filter : str, optional (default: None)
        If set to a metabolic model's Mongo _id, only get spectra for compounds
        in or derived from that metabolic model.
    putative : bool, optional (default: True)
        If False, only find known compounds (i.e. in Generation 0). Otherwise,
        finds both known and predicted compounds.

    Returns
    -------
    spectral_library : str
        Text of all matching spectra, including headers and peak lists.
    """

    def print_peaklist(peaklist):
        text = [f"Num Peaks: {len(peaklist)}"]
        for x in peaklist:
            text.append(f"{x[0]} {x[1]}")
        text.append("")
        return text

    spectral_library = []
    msp_projection = {
        "MINE_id": 1,
        "KEGG_id": 1,
        "Mass": 1,
        "Inchikey": 1,
        "Formula": 1,
        "SMILES": 1,
        "logP": 1,
        "Charge": 1,
        "Spectra.Positive": 1,
        "Spectra.Negative": 1,
    }

    query_dict = {"_id": mongo_id}

    compound = db.compounds.find_one(query_dict, msp_projection)

    # add header
    header = []
    header.append(f"Name: MINE Compound {compound['_id']}")

    for k, v in compound.items():
        if k not in {"_id", "Spectra"}:
            header.append(f"{k}: {v}")

    header.append("Instrument: CFM-ID 4.0")

    # add peak lists
    if compound["Spectra"]:
        for ion_mode in ["Positive", "Negative"]:
            for energy, spec in compound["Spectra"][ion_mode].items():
                spectral_library += header
                spectral_library += [f"Ionization: {ion_mode}", f"Energy: {energy}"]
                spectral_library += print_peaklist(spec)

    spectral_library = "\n".join(spectral_library)

    return spectral_library


def score_compounds(
    compounds: list,
    model_id: str = None,
    core_db: pymongo.database = None,
    mine_db: pymongo.database = None,
    kegg_db: pymongo.database = None,
    parent_frac: float = 0.75,
    reaction_frac: float = 0.25,
    get_native: bool = False,
) -> List[dict]:
    """This function validates compounds against a metabolic model, returning
    only the compounds which pass.

    Parameters
    ----------
    db : Mongo DB
        Should contain a "models" collection with compound and reaction IDs
        listed.
    core_db : Mongo DB
        Core MINE database.
    compounds : list
        Each element is a dict describing that compound. Should have an '_id'
        field.
    model_id : str
        KEGG organism code (e.g. 'hsa').
    parent_frac : float, optional
        Weighting for compounds derived from compounds in the provided model.
        0.75 by default.
    reaction_frac : float, optional
        Weighting for compounds derived from known compounds not in the model.
        0.25 by default.

    Returns
    -------
    compounds : List[dict]
        Modified version of input compounds list, where each compound now has
        a 'Likelihood_score' key and value between 0 and 1.
    """
    for cpd in compounds:
        cpd["native_hit"] = False
        cpd["product_of_native_hit"] = False

    if get_native and core_db and mine_db and kegg_db and model_id:
        cpd_ids = [cpd["_id"] for cpd in compounds]
        met_dataset = MetabolomicsDataset(name="temp")
        met_dataset.native_set = get_KEGG_comps(mine_db, core_db, kegg_db, [model_id])
        native_product_ids = met_dataset.check_product_of_native(cpd_ids, mine_db)

        for cpd in compounds:
            if cpd["_id"] in met_dataset.native_set:
                cpd["native_hit"] = True
            elif cpd["_id"] in native_product_ids:
                cpd["product_of_native_hit"] = True

    for comp in compounds:
        if comp["native_hit"]:
            comp["Likelihood_score"] = parent_frac + reaction_frac
        if comp["product_of_native_hit"]:
            comp["Likelihood_score"] = parent_frac

    return compounds



"""Tests for databases.py using pytest."""
# pylint: disable=redefined-outer-name
# pylint: disable=protected-access

import json
import os
import shutil
from copy import copy, deepcopy
from pathlib import Path
from shutil import rmtree

import pymongo
import pytest
from pymongo.errors import ServerSelectionTimeoutError

from minedatabase import utils
from minedatabase.databases import (
    MINE,
    write_compounds_to_mine,
    write_core_compounds,
    write_reactions_to_mine,
)


try:
    client = pymongo.MongoClient(ServerSelectionTimeoutMS=10)
    client.server_info()
    del client
except ServerSelectionTimeoutError as err:
    pytest.skip("No Mongo Server Found", allow_module_level=True)

file_path = Path(__file__)
file_dir = file_path.parent

DATA_DIR = (file_dir / "../data/").resolve()


@pytest.fixture()
def cpd_dict():
    """Create a compound dict for testing."""
    cpd_dict = {
        "_id": "Ctest",
        "SMILES": (
            "Nc1ncnc2c1ncn2[C@@H]1O[C@H](COP(=O)(O)OS(=O)(=O)O)[C@@H]"
            "(OP(=O)(O)O)[C@H]1O"
        ),
        "Inchi": (
            "InChI=1S/C10H15N5O13P2S/c11-8-5-9(13-2-12-8)15(3-14-5)10-"
            "6(16)7(27-29(17,18)19)4(26-10)1-25-30(20,21)28-31(22,23)"
            "24/h2-4,6-7,10,16H,1H2,(H,20,21)(H2,11,12,13)(H2,17,18,19"
            ")(H,22,23,24)/t4-,6-,7-,10-/m1/s1"
        ),
        "Type": "Coreactant",
        "Generation": 0,
        "Formula": "C10H15N5O13P2S",
        "Expand": False,
    }

    yield cpd_dict


@pytest.fixture()
def rxn_dicts():
    """Create two reaction dictionaries for testing."""
    rxn_dicts = [
        {
            "_id": "RXN1",
            "Reactants": [[1, "cpd1"], [2, "cpd2"]],
            "Products": [[1, "cpd3"], [1, "cpd4"]],
            "Operators": ["op1", "op2"],
            "SMILES_rxn": "(1) A + (2) B -> (1) C + (2) D",
        },
        {
            "_id": "RXN2",
            "Reactants": [[1, "cpd1"], [2, "cpd2"]],
            "Products": [[1, "cpd3"], [1, "cpd4"]],
            "Operators": ["op1"],
            "SMILES_rxn": "(1) A + (2) B -> (1) C + (2) D",
        },
    ]
    return rxn_dicts


@pytest.mark.skipif(
    os.name == "nt", reason="MolConvert fails on Windows due to permissions errors"
)
def test_generate_image_files(test_db):
    """Test image generation."""
    img_dir = DATA_DIR / "imgs/"
    test_db.generate_image_files(img_dir)
    try:
        assert (img_dir / "C455bc3dc93cd3bb3bef92a34767693a4716aa3fb.svg").is_file()
        assert len(os.listdir(img_dir)) == 26
    finally:
        rmtree(img_dir)
    # Use 'Generation': 1 to do this for just the starting compound
    test_db.generate_image_files(img_dir, {"Generation": 1}, 3, "png")
    try:
        assert (
            img_dir / "C/c/f" / "Ccffda1b2e82fcdb0e1e710cad4d5f70df7a5d74f.png"
        ).is_file()
    finally:
        shutil.rmtree(img_dir)


def test_insert_single_mine_compound(test_db):
    """Test single mine insertion."""
    smiles = "CC(=O)O"
    compound_dict = [{"_id": "test_cpd", "SMILES": smiles}]
    write_compounds_to_mine(compound_dict, test_db)

    entry = test_db.compounds.find_one({"SMILES": smiles})
    assert entry
    assert entry["_id"] == "test_cpd"


def test_insert_bulk_mine_compounds(test_db):
    """Test inserting bulk compounds."""

    smiles1 = "CC(=O)O"
    smiles2 = "CCN"
    compound_dict = [
        {"_id": "test_cpd1", "SMILES": smiles1},
        {"_id": "test_cpd2", "SMILES": smiles2},
    ]

    write_compounds_to_mine(compound_dict, test_db)

    entry1 = test_db.compounds.find_one({"SMILES": smiles1})
    assert entry1
    assert entry1["_id"] == "test_cpd"

    entry2 = test_db.compounds.find_one({"SMILES": smiles2})
    assert entry2
    assert entry2["_id"] == "test_cpd2"


def test_insert_single_core_compound(test_db, cpd_dict):
    """Test inserting a single core compound."""
    write_core_compounds([cpd_dict], test_db, "test")

    try:
        entry = test_db.core_compounds.find_one({"_id": cpd_dict["_id"]})
        assert entry
        assert isinstance(entry["Mass"], float)
        assert entry["Inchi"]
        assert entry["Inchikey"]
    finally:
        test_db.core_compounds.delete_many({"_id": cpd_dict["_id"]})


def test_insert_bulk_core_compound(test_db, cpd_dict):
    """Test inserting many core compounds."""

    cpd_dict2 = copy(cpd_dict)
    cpd_dict2["_id"] = "cpd_dict2"

    write_core_compounds([cpd_dict, cpd_dict2], test_db, "test")
    #
    try:
        for smiles in [cpd_dict["SMILES"], cpd_dict2["SMILES"]]:
            entry = test_db.core_compounds.find_one({"SMILES": smiles})
            assert entry
            assert isinstance(entry["Mass"], float)
            assert entry["Inchi"]
            assert entry["Inchikey"]
    finally:
        for i in [0, 1]:
            test_db.core_compounds.delete_many({"_id": f"test_mine_cpd{i}"})


def test_write_reaction(test_db, rxn_dicts):
    """Test writing reactions."""
    write_reactions_to_mine(rxn_dicts, test_db)

    assert (deepcopy(rxn_dicts[0])) == deepcopy(
        test_db.reactions.find_one({"_id": "RXN1"})
    )
    assert (deepcopy(rxn_dicts[1])) == deepcopy(
        test_db.reactions.find_one({"_id": "RXN2"})
    )


def db_driver(test_db):
    """Test database driver."""
    assert isinstance(test_db.compounds, pymongo.collection.Collection)
    assert isinstance(test_db.reactions, pymongo.collection.Collection)
    assert isinstance(test_db.operators, pymongo.collection.Collection)
    assert isinstance(test_db.core_compounds, pymongo.collection.Collection)
    assert isinstance(test_db._db, pymongo.database.Database)



"""Reaction.py: Methods to execute reactions."""
import collections
import multiprocessing
from functools import partial
from typing import Tuple

import rdkit.rdBase as rkrb
import rdkit.RDLogger as rkl
from rdkit.Chem.AllChem import (
    AddHs,
    CalcMolFormula,
    Kekulize,
    MolFromSmiles,
    MolToSmiles,
    RDKFingerprint,
    RemoveHs,
    SanitizeMol,
    GetFormalCharge,
)

from minedatabase import utils


logger = rkl.logger()
logger.setLevel(rkl.ERROR)
rkrb.DisableLog("rdApp.error")

###############################################################################
# Functions to run transformations
# There are two distinct functions to transform two flavors of operators.
#   1. Full Operators are the operators as loaded directly from the list
#       of operatores. These operators use a single supplied molecule for all
#       "Any" instances in the rule.
#   2. Partial operators are operators for reactions with more than one
#       "Any" in the reactants. These rules are derived from individually
#       mapped reactions and are called partial because only one "Any" is
#       allowed to be novel, the other "Any"s are determined by the
#       mapped reactions.

# Both operators are preprocessed slightly differently, but yield the same
# output format back to the pickaxe object.


def _run_reaction(
    rule_name: str,
    rule: tuple,
    reactant_mols: dict,
    coreactant_mols: dict,
    coreactant_dict: dict,
    local_cpds: dict,
    local_rxns: dict,
    generation: int,
    explicit_h: bool,
) -> Tuple[dict, dict]:
    """Apply reaction rules to a list of compounds.

    Transform list of mols and a reaction rule into a half reaction
    describing either the reactants or products.

    Parameters
    ----------
    rule_name : str
        Name of the rule to run.
    rule : tuple
        Rule to run.
    reactant_mols : dict
        Reactant molecules for reaction.
    coreactant_mols : dict
        Coreactants for the reaction.
    coreactant_dict : dict
        Dictionary of coreactants.
    local_cpds : dict
        A dictionary of compounds generated by reactions prior.
    local_rxns : dict
        A dictionary of reactions already generated.
    generation : int
        Which generation this reaction is.
    explicit_h : bool
        Whether or not to explicitly use hydrogens.

    Returns
    -------
    Tuple[dict, dict]
        A tuple of local compounds, local reactions.
    """

    def _make_half_rxn(mol_list, rules):
        cpds = {}
        cpd_counter = collections.Counter()

        # correct for number of H based on charge
        charge_correction = 0

        for mol, rule in zip(mol_list, rules):
            if rule == "Any":
                cpd_dict = _gen_compound(mol)
                # failed compound
                if cpd_dict is None:
                    return None, None
            else:
                cpd_id = coreactant_mols[rule][1]
                cpd_dict = coreactant_dict[cpd_id]

            cpds[cpd_dict["_id"]] = cpd_dict
            cpd_counter.update({cpd_dict["_id"]: 1})
            charge_correction += GetFormalCharge(mol)

        atom_counts = collections.Counter()
        for cpd_id, cpd_dict in cpds.items():
            for atom_id, atom_count in cpd_dict["atom_count"].items():
                atom_counts[atom_id] += atom_count * cpd_counter[cpd_id]

        # correct for number of H based on charge
        atom_counts["H"] -= charge_correction

        cpd_returns = [(stoich, cpds[cpd_id]) for cpd_id, stoich in cpd_counter.items()]

        return cpd_returns, atom_counts

    def _gen_compound(mol):
        rkl.DisableLog("rdApp.*")
        try:
            if explicit_h:
                mol = RemoveHs(mol)

            # resolve potential tautomers and choose first one
            mol_smiles = MolToSmiles(mol, True)
            if "n" in mol_smiles:
                mol_smiles = utils.postsanitize_smiles([mol_smiles])[0][0]
                mol = MolFromSmiles(mol_smiles)

            SanitizeMol(mol)

        # TODO: logger
        # Get lots of "Explicit valence greater than permitted" errors here
        # This is for predicted compounds that are infeasible, so we throw them out
        except BaseException:
            return None
        rkl.EnableLog("rdApp.*")

        mol_smiles = MolToSmiles(mol, True)
        if "." in mol_smiles:
            return None

        cpd_id, inchi_key = utils.get_compound_hash(mol_smiles, "Predicted")
        if cpd_id:
            if cpd_id not in local_cpds:
                cpd_dict = {
                    "ID": None,
                    "_id": cpd_id,
                    "SMILES": mol_smiles,
                    "InChI_key": inchi_key,
                    "Type": "Predicted",
                    "Generation": generation,
                    "atom_count": utils.get_atom_count(mol),
                    "Reactant_in": [],
                    "Product_of": [],
                    "Expand": True,
                    "Formula": CalcMolFormula(mol),
                    "last_tani": 0,
                }
            else:
                cpd_dict = local_cpds[cpd_id]

            return cpd_dict
        else:
            return None

    try:
        product_sets = rule[0].RunReactants(reactant_mols, maxProducts=10000)
        reactants, reactant_atoms = _make_half_rxn(reactant_mols, rule[1]["Reactants"])
    except BaseException:
        reactants = None, None

    if reactants is None:
        reactants = None, None

    if not all(reactants):
        return local_cpds, local_rxns

    reactant_set = set([r[1]["_id"] for r in reactants])

    for product_mols in product_sets:
        try:
            products, product_atoms = _make_half_rxn(product_mols, rule[1]["Products"])
            if not products:
                continue

            # Check to see if any reactants weren't changed
            product_set = set([p[1]["_id"] for p in products])
            if reactant_set.intersection(product_set):
                continue

            if reactant_atoms - product_atoms or product_atoms - reactant_atoms:
                is_atom_balanced = False
            else:
                is_atom_balanced = True

            if is_atom_balanced:
                for _, cpd_dict in products:
                    if cpd_dict["_id"].startswith("C"):
                        local_cpds.update({cpd_dict["_id"]: cpd_dict})

                rhash, rxn_text = utils.get_reaction_hash(reactants, products)
                if rhash not in local_rxns:
                    local_rxns[rhash] = {
                        "_id": rhash,
                        # give stoich and id of reactants/products
                        "Reactants": [(s, r["_id"]) for s, r in reactants],
                        "Products": [(s, p["_id"]) for s, p in products],
                        "Operators": {rule_name},
                        "SMILES_rxn": rxn_text,
                    }
                else:
                    local_rxns[rhash]["Operators"].add(rule_name)
        # TODO: Logger
        except BaseException:
            continue
    # return compounds and reactions to be added into the local
    return local_cpds, local_rxns


# Full Operators
def _transform_ind_compound_with_full(
    coreactant_mols: dict,
    coreactant_dict: dict,
    operators: list,
    generation: int,
    explicit_h: bool,
    kekulize: bool,
    compound_smiles: list,
) -> Tuple[dict, dict]:
    """Transform a compound with the full reaction operators.

    Apply the reaction rules as written to a compound.

    Parameters
    ----------
    coreactant_mols : dict
        Coreactant molecules.
    coreactant_dict : dict
        Coreactant dictionary.
    operators : list
        Reaction operators.
    generation : int
        Which generation these reactions are in.
    explicit_h : bool
        Whether or not to use explicit hydrogen.
    kekulize : bool
        Whether or not to kekulize molecules.
    compound_smiles : list
        List of compound smiles.

    Returns
    -------
    Tuple[dict, dict]
        Tuple of local compounds, local reactions.
    """
    local_cpds = dict()
    local_rxns = dict()

    mol = MolFromSmiles(compound_smiles)
    mol = RemoveHs(mol)
    if not mol:
        print(f"Unable to parse: {compound_smiles}")
        return None
    if kekulize:
        Kekulize(mol, clearAromaticFlags=True)
    if explicit_h:
        mol = AddHs(mol)
    # Apply reaction rules to prepared compound

    # run through the single compound operatores
    for rule_name, rule in operators.items():
        # Get RDKit Mol objects for reactants
        reactant_mols = tuple(
            [mol if x == "Any" else coreactant_mols[x][0] for x in rule[1]["Reactants"]]
        )
        # Perform chemical reaction on reactants for each rule
        # try:
        generated_cpds, generated_rxns = _run_reaction(
            rule_name,
            rule,
            reactant_mols,
            coreactant_mols,
            coreactant_dict,
            local_cpds,
            local_rxns,
            generation,
            explicit_h,
        )

        local_cpds.update(generated_cpds)
        for rxn, vals in generated_rxns.items():
            if rxn in local_rxns:
                local_rxns[rxn]["Operators"].union(vals["Operators"])

    return local_cpds, local_rxns


def transform_all_compounds_with_full(
    compound_smiles: list,
    coreactants: dict,
    coreactant_dict: dict,
    operators: dict,
    generation: int,
    explicit_h: bool,
    kekulize: bool,
    processes: int,
) -> Tuple[dict, dict]:
    """Transform compounds given a list of rules.

    Carry out the transformation of a list of compounds given operators.
    Generates new products and returns them to be processed by pickaxe.

    Parameters
    ----------
    compound_smiles : list
        List of SMILES to react.
    coreactants : dict
        Dictionary of correactants RDKit Mols defined in rules.
    coreactant_dict : dict
        Dictionary of correactant compoudnds defined in rules.
    operators : dict
        Dictionary of reaction rules.
    generation : int
        Value of generation to expand.
    explicit_h : bool
        Whether or not to have explicit Hs in reactions.
    kekulize : bool
        Whether or not to kekulize compounds.
    processes : int
        Number of processors being used.

    Returns
    -------
    Tuple[dict, dict]
        Returns a tuple of New Compounds and New Reactants.
    """

    def print_progress(done, total):
        # Use print_on to print % completion roughly every 2.5 percent
        # Include max to print no more than once per compound (e.g. if
        # less than 20 compounds)
        print_on = max(round(0.1 * total), 1)
        if not done % print_on:
            print(
                f"Generation {generation}: {round(done / total * 100)}"
                " percent complete"
            )

    # First transform
    new_cpds_master = {}
    new_rxns_master = {}

    transform_compound_partial = partial(
        _transform_ind_compound_with_full,
        coreactants,
        coreactant_dict,
        operators,
        generation,
        explicit_h,
        kekulize,
    )
    # par loop
    if processes > 1:
        chunk_size = max([round(len(compound_smiles) / (processes * 10)), 1])
        pool = multiprocessing.Pool(processes=processes)
        for i, res in enumerate(
            pool.imap_unordered(transform_compound_partial, compound_smiles, chunk_size)
        ):
            new_cpds, new_rxns = res
            new_cpds_master.update(new_cpds)

            # Need to check if reactions already exist to update operators list
            for rxn, rxn_dict in new_rxns.items():
                if rxn in new_rxns_master:
                    ops_set = rxn_dict["Operators"]
                    new_rxns_master[rxn]["Operators"].union(ops_set)
                else:
                    new_rxns_master.update({rxn: rxn_dict})
            print_progress(i, len(compound_smiles))

    else:
        for i, smiles in enumerate(compound_smiles):
            new_cpds, new_rxns = transform_compound_partial(smiles)
            # new_cpds as cpd_id:cpd_dict
            # new_rxns as rxn_id:rxn_dict
            new_cpds_master.update(new_cpds)
            # Need to check if reactions already exist to update operators list
            for rxn, rxn_dict in new_rxns.items():
                if rxn in new_rxns_master:
                    ops_set = rxn_dict["Operators"]
                    new_rxns_master[rxn]["Operators"].union(ops_set)
                else:
                    new_rxns_master.update({rxn: rxn_dict})
            print_progress(i, len(compound_smiles))

    return new_cpds_master, new_rxns_master


# TODO: Partial operators aren't used... keep here or move?
# Partial Operators
# def _transform_ind_compound_with_partial(
#     coreactant_mols: dict,
#     coreactant_dict: dict,
#     operators: dict,
#     generation: int,
#     explicit_h: bool,
#     partial_rules: dict,
#     compound_smiles: list,
# ):
#     # 1. See if rule matches the compound passed
#     #   (rule from partial_rules dict keys)
#     # 2. If match apply transform_ind_compound_with_full to each
#     def generate_partial_mols(partial_rule):
#         def gen_mol(smi):
#             mol = MolFromSmiles(smi)
#             mol = RemoveHs(mol)
#             Kekulize(mol, clearAromaticFlags=True)
#             if explicit_h:
#                 mol = AddHs(mol)
#             return mol

#         rule_reactants = operators[partial_rule["rule"]][1]["Reactants"]
#         cofactor = [False if r == "Any" else True for r in rule_reactants]
#         reactant_mols = []
#         for is_cofactor, smi in zip(cofactor, partial_rule["reactants"]):
#             if is_cofactor:
#                 reactant_mols.append(coreactant_mols[smi][0])
#             elif smi == "SMARTS_match":
#                 reactant_mols.append(gen_mol(compound_smiles))
#             else:
#                 # These reactions already happen with any;any
# if (
#     utils.get_compound_hash(smi)
#     != utils.get_compound_hash(compound_smiles)
# ):
#                     reactant_mols.append(gen_mol(smi))
#                 else:
#                     return None
#         return reactant_mols

#     local_cpds = dict()
#     local_rxns = dict()

#     mol = MolFromSmiles(compound_smiles)
#     mol = RemoveHs(mol)
#     if not mol:
#         print(f"Unable to parse: {compound_smiles}")
#         return None
#     Kekulize(mol, clearAromaticFlags=True)
#     if explicit_h:
#         mol = AddHs(mol)
#     # Apply reaction rules to prepared compound

#     # run through the single compound operatores
#     for ind_SMARTS, rules in partial_rules.items():
#         # does mol match vs smiles match change things?
#         if utils.QuickSmartsMatch(compound_smiles, ind_SMARTS):
#             for partial_rule in rules:
#                 # Perform chemical reaction on reactants for each rule
#                 # try:
#                 rule_name = partial_rule["rule_reaction"].split("_")[0]
#                 rule = operators[partial_rule["rule"]]
#                 reactant_mols = generate_partial_mols(partial_rule)
#                 if reactant_mols:
#                     generated_cpds, generated_rxns = _run_reaction(
#                         rule_name,
#                         rule,
#                         reactant_mols,
#                         coreactant_mols,
#                         coreactant_dict,
#                         local_cpds,
#                         local_rxns,
#                         generation,
#                         explicit_h,
#                     )

#                     local_cpds.update(generated_cpds)
#                     for rxn, vals in generated_rxns.items():
#                         if rxn in local_rxns:
#                             if "Partial Operators" in local_rxns[rxn]:
#                                 local_rxns[rxn]["Partial Operators"].update(
#                                     [partial_rule["rule_reaction"]]
#                                 )
#                             else:
#                                 local_rxns[rxn]["Partial Operators"] = set(
#                                     [partial_rule["rule_reaction"]]
#                                 )
#     return local_cpds, local_rxns


# def _transform_all_compounds_with_partial(
#     compound_smiles,
#     coreactants,
#     coreactant_dict,
#     operators,
#     generation,
#     explicit_h,
#     processes,
#     partial_rules,
# ):
#     """
#     Stage partial operator reactions.

#     This function accepts in a list of cpds (cpd_list) and runs the
#     transformation in parallel of these.
#     """

#     def print_progress(done, total):
#         # Use print_on to print % completion roughly every 2.5 percent
#         # Include max to print no more than once per compound (e.g. if
#         # less than 20 compounds)
#         print_on = max(round(0.1 * total), 1)
#         if not done % print_on:
#             print(
#                 f"Generation {generation}: {round(done / total * 100)}"
#                 " percent complete"
#             )

#     # First transform
#     new_cpds_master = {}
#     new_rxns_master = {}

#     transform_compound_partial = partial(
#         _transform_ind_compound_with_partial,
#         coreactants,
#         coreactant_dict,
#         operators,
#         generation,
#         explicit_h,
#         partial_rules,
#     )
#     # par loop
#     if processes > 1:
#         chunk_size = 1
#         # print(f'Chunk size = {chunk_size}')
#         pool = multiprocessing.Pool(processes=processes)
#         for i, res in enumerate(
#             pool.imap_unordered(transform_compound_partial, compound_smiles,
#               chunk_size)
#         ):
#             new_cpds, new_rxns = res
#             new_cpds_master.update(new_cpds)

#             # Need to check if reactions already exist to update operators list
#             for rxn, rxn_dict in new_rxns.items():
#                 if rxn in new_rxns_master:
#                     new_rxns_master[rxn]["Operators"].union(rxn_dict["Operators"])
#                 else:
#                     new_rxns_master.update({rxn: rxn_dict})
#             print_progress(i, len(compound_smiles))

#     else:
#         for i, smiles in enumerate(compound_smiles):
#             new_cpds, new_rxns = transform_compound_partial(smiles)
#             # new_cpds as cpd_id:cpd_dict
#             # new_rxns as rxn_id:rxn_dict
#             new_cpds_master.update(new_cpds)
#             # Need to check if reactions already exist to update operators list
#             for rxn, rxn_dict in new_rxns.items():
#                 if rxn in new_rxns_master:
#                     new_rxns_master[rxn]["Partial Operators"] = new_rxns_master[rxn][
#                         "Partial Operators"
#                     ].union(rxn_dict["Partial Operators"])
#                 else:
#                     new_rxns_master.update({rxn: rxn_dict})
#             print_progress(i, len(compound_smiles))

#     return new_cpds_master, new_rxns_master
